# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Tools related to binary linking.
"""

from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from typing import cast, Literal

from typecode import magic2  # type: ignore[import-untyped]

logger = logging.getLogger(__name__)
del logging


ELF_EXE = "executable"
ELF_SHARED = "shared object"
ELF_RELOC = "relocatable"
ELF_UNKNOWN = "unknown"
ELF_TYPES = [ELF_EXE, ELF_SHARED, ELF_RELOC]
ELF_TYPES_TYPE = Literal["executable", "shared object", "relocatable", "unknown"]


def _get_file_type(path: Path) -> str:
    """
    Get the file type.

    :param: The file to check.
    :return: The guessed file type.
    """
    return cast(str, magic2.file_type(path))


def is_elf(path: Path) -> bool:
    """
    Check if the given file is an ELF file.

    :param path: The file to check.
    :return: True if the file is an ELF binary, False otherwise.
    """
    file_type = _get_file_type(path).lower()
    return file_type.startswith("elf") and any(elf_type in file_type for elf_type in ELF_TYPES)


def get_elf_type(path: Path) -> ELF_TYPES_TYPE | None:
    """
    Get the ELF type of the given file.

    :param path: The file to check.
    :return: The ELF type of the given file if it is an ELF binary, None otherwise.
    """
    if not is_elf(path):
        return None
    file_type = _get_file_type(path).lower()
    for elf_type in ELF_TYPES:
        if elf_type in file_type:
            return cast(ELF_TYPES_TYPE, elf_type)
    return cast(ELF_TYPES_TYPE, ELF_UNKNOWN)


def check_shared_objects(path: Path) -> str | None:
    """
    Check which other shared objects a shared object links to.

    :param path: The file path to analyze.
    :return: The analysis results if the path points to a shared object, `None` otherwise.
    """
    if not is_elf(path):
        return None
    if path.is_symlink():
        # Ignore symlinks as they usually are package-internal and `ldd` does not always like them.
        logger.warning(
            "Ignoring symlink %s to %s for shared object analysis.", path, path.resolve()
        )
        return None
    output = subprocess.check_output(["ldd", path], stderr=subprocess.PIPE)
    return output.decode("UTF-8")



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Retrieve license-related data.
"""

from __future__ import annotations

import atexit
import math
import shutil
from collections import defaultdict
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import BinaryIO, cast, Generator

import scancode_config  # type: ignore[import-untyped]
from joblib import Parallel, delayed  # type: ignore[import-untyped]

from license_tools.tools import cargo_tools, font_tools, linking_tools, pip_tools, scancode_tools
from license_tools.tools.pip_tools import download_package
from license_tools.tools.scancode_tools import FileResults, Licenses, PackageResults
from license_tools.utils import archive_utils
from license_tools.utils.download_utils import download_file
from license_tools.utils.path_utils import DirectoryWithFixedNameContext, get_files_from_directory


class RetrievalFlags:
    """
    Data retrieval flags to get shorter parameter lists.
    """

    COPYRIGHTS = 1
    EMAILS = 2
    FILE_INFO = 4
    URLS = 8
    LDD_DATA = 16
    FONT_DATA = 32
    PYTHON_METADATA = 64
    CARGO_METADATA = 128

    @classmethod
    def to_int(
        cls,
        retrieve_copyrights: bool = False,
        retrieve_emails: bool = False,
        retrieve_file_info: bool = False,
        retrieve_urls: bool = False,
        retrieve_ldd_data: bool = False,
        retrieve_font_data: bool = False,
        retrieve_python_metadata: bool = False,
        retrieve_cargo_metadata: bool = False,
    ) -> int:
        """
        Convert the given boolean parameter values to a single integer flag value.

        :param retrieve_copyrights: Whether to retrieve copyright information.
        :param retrieve_emails: Whether to retrieve e-mails.
        :param retrieve_file_info: Whether to retrieve file-specific information.
        :param retrieve_urls: Whether to retrieve URLs.
        :param retrieve_ldd_data: Whether to retrieve linking data for shared objects.
        :param retrieve_font_data: Whether to retrieve font data.
        :param retrieve_python_metadata: Whether to retrieve Python package metadata.
        :param retrieve_cargo_metadata: Whether to retrieve Cargo metadata.
        :return: The flags derived from the given parameters.
        """
        return (
            cls.COPYRIGHTS * retrieve_copyrights
            + cls.EMAILS * retrieve_emails
            + cls.FILE_INFO * retrieve_file_info
            + cls.URLS * retrieve_urls
            + cls.LDD_DATA * retrieve_ldd_data
            + cls.FONT_DATA * retrieve_font_data
            + cls.PYTHON_METADATA * retrieve_python_metadata
            + cls.CARGO_METADATA * retrieve_cargo_metadata
        )

    @classmethod
    def all(cls, as_kwargs: bool = False) -> int | dict[str, bool]:
        """
        Utility method to enable all flags.

        :param: If enabled, return kwargs instead of the integer value.
        :return: The value for all flags enabled.
        """
        value = cls.to_int(True, True, True, True, True, True, True, True)
        if as_kwargs:
            return cls.to_kwargs(value)
        return value

    @classmethod
    def is_set(cls, flags: int, flag: int) -> bool:
        """
        Check if the given flag is set.

        :param flags: The flags to check inside.
        :param flag: The flag value to check for.
        :return: The check result.
        """
        return bool(flags & flag)

    @classmethod
    def to_kwargs(cls, flags: int) -> dict[str, bool]:
        """
        Convert the given flags to keyword arguments.

        :param flags: The flags to convert.
        :return: The associated keyword arguments.
        """
        return dict(
            retrieve_copyrights=cls.is_set(flags=flags, flag=cls.COPYRIGHTS),
            retrieve_emails=cls.is_set(flags=flags, flag=cls.EMAILS),
            retrieve_file_info=cls.is_set(flags=flags, flag=cls.FILE_INFO),
            retrieve_urls=cls.is_set(flags=flags, flag=cls.URLS),
            retrieve_ldd_data=cls.is_set(flags=flags, flag=cls.LDD_DATA),
            retrieve_font_data=cls.is_set(flags=flags, flag=cls.FONT_DATA),
            retrieve_python_metadata=cls.is_set(flags=flags, flag=cls.PYTHON_METADATA),
            retrieve_cargo_metadata=cls.is_set(flags=flags, flag=cls.CARGO_METADATA),
        )


def _run_on_archive_file(path: Path, short_path: str, default_to_none: bool = False) -> FileResults | None:
    """
    Run archive-specific analysis.

    :param path: The path to run on.
    :param short_path: The short path to use for display.
    :param default_to_none: Return `None` if no results could be determined instead of an empty entry.
    :return: Explicitly declared results.
    """

    if path.suffix == ".rpm" or (path.suffixes and ".rpm" in path.suffixes):
        rpm_results = PackageResults.from_rpm(path)
        if rpm_results.declared_license_expression_spdx:
            results = _get_dummy_file_results(path=path, short_path=short_path)
            results.licenses = Licenses(
                detected_license_expression=rpm_results.declared_license_expression_spdx,
                detected_license_expression_spdx=rpm_results.declared_license_expression_spdx,
            )
            return results
    if default_to_none:
        return None
    return _get_dummy_file_results(path=path, short_path=short_path)


def _get_dummy_file_results(
        path: Path,
        short_path: str
) -> FileResults:
    """
    Get some empty dummy license results, which allows displaying the file in the
    results, but skipping the actual analysis.

    :param path: The file path to analyze.
    :param short_path: The short path to use for display.
    :return: Minimal results.
    """
    results = FileResults(path=path, short_path=short_path)
    results.licenses = Licenses()
    return results


def run_on_file(
    path: Path,
    short_path: str,
    retrieval_flags: int = 0,
) -> FileResults:
    """
    Run the analysis on the given file.

    :param path: The file path to analyze.
    :param short_path: The short path to use for display.
    :param retrieval_flags: Values to retrieve.
    :return: The requested results.
    """
    if archive_utils.can_extract(archive_path=path):
        # Archive files which can be extracted further are not being analyzed on the
        # file level. This should improve the extraction speed and avoid possible
        # memory errors as this skips running string matching on possibly large
        # archives which are binary blobs anyway and do not provide any real value.
        # Instead, just have a quick look at their headers if they provide any useful
        # values.
        return cast(
            FileResults,
            _run_on_archive_file(path=path, short_path=short_path, default_to_none=False)
        )

    retrieval_kwargs = RetrievalFlags.to_kwargs(flags=retrieval_flags)

    if path.name.startswith("Cargo.toml") and retrieval_kwargs.pop("retrieve_cargo_metadata"):
        print(short_path)
        print(cargo_tools.check_metadata(path=path) + "\n")

    # This data is not yet part of the dataclasses above, as it is a custom analysis.
    # Return early if we got a result here, as these binary files currently do not
    # provide any additional useful insights in most of the cases, but tend to be
    # larger binary blobs which just slow down the analysis.
    if retrieval_kwargs.pop("retrieve_ldd_data"):
        results = linking_tools.check_shared_objects(path=path)
        if results:
            print(short_path + "\n" + results)
            return _get_dummy_file_results(path=path, short_path=short_path)
    if retrieval_kwargs.pop("retrieve_font_data"):
        results = font_tools.check_font(path=path)
        if results:
            print(short_path + "\n" + results + "\n")
            return _get_dummy_file_results(path=path, short_path=short_path)

    # Register this here as each parallel process has its own directory.
    atexit.register(scancode_tools.cleanup, scancode_config.scancode_temp_dir)

    return FileResults(
        path=path, short_path=short_path, retrieve_licenses=True,
        retrieve_copyrights=retrieval_kwargs["retrieve_copyrights"],
        retrieve_emails=retrieval_kwargs["retrieve_emails"],
        retrieve_urls=retrieval_kwargs["retrieve_urls"],
        retrieve_file_info=retrieval_kwargs["retrieve_file_info"],
    )


def run_on_directory(
    directory: str,
    job_count: int = 4,
    retrieval_flags: int = 0,
    prefix: str | None = None,
    allow_random_directory_for_archive: bool = True,
    delete_unpacked_archive_directories: bool = True,
) -> Generator[FileResults, None, None]:
    """
    Run the analysis on the given directory.

    :param directory: The directory to analyze.
    :param job_count: The number of parallel jobs to use.
    :param retrieval_flags: Values to retrieve.
    :param prefix: Custom prefix to use.
    :param allow_random_directory_for_archive: Allow a random directory for unpacking archives.
    :param delete_unpacked_archive_directories: Delete the directories of unpacked archives afterwards.
    :return: The requested results per file.
    """
    files = list(get_files_from_directory(directory, prefix))
    results = Parallel(n_jobs=job_count)(
        delayed(run_on_file)(
            path=path,
            short_path=short_path,
            retrieval_flags=retrieval_flags,
        )
        for path, short_path in files
    )
    yield from results

    for path, _ in files:
        if archive_utils.can_extract(path):
            name = path.name[:-len("".join(path.suffixes))]
            subdirectory = path.parent / f'{name}_{"_".join(path.suffixes).replace(".", "")}'
            with DirectoryWithFixedNameContext(
                    directory=subdirectory,
                    fallback_to_random_if_exists=allow_random_directory_for_archive,
                    delete_afterwards=delete_unpacked_archive_directories,
            ) as target_directory:
                archive_utils.extract(archive_path=path, target_directory=target_directory)
                yield from run_on_directory(
                    directory=str(target_directory),
                    job_count=job_count,
                    retrieval_flags=retrieval_flags,
                    prefix=directory,
                )


def run_on_package_archive_file(
    archive_path: Path,
    job_count: int = 4,
    retrieval_flags: int = 0,
) -> Generator[FileResults, None, None]:
    """
    Run the analysis on the given package archive file.

    :param archive_path: The package archive path to analyze.
    :param job_count: The number of parallel jobs to use.
    :param retrieval_flags: Values to retrieve.
    :return: The requested results.
    """
    archive_results = _run_on_archive_file(path=archive_path, short_path=archive_path.name, default_to_none=True)
    if archive_results is not None:
        yield archive_results

    with TemporaryDirectory() as working_directory:
        if not archive_utils.can_extract(archive_path):
            raise ValueError(f'Unsupported archive format: {archive_path}')
        archive_utils.extract(archive_path=archive_path, target_directory=Path(working_directory))

        if RetrievalFlags.is_set(flags=retrieval_flags, flag=RetrievalFlags.PYTHON_METADATA):
            print(pip_tools.check_metadata(path=working_directory) + "\n")

        yield from run_on_directory(
            directory=working_directory,
            job_count=job_count,
            retrieval_flags=retrieval_flags,
        )


def run_on_downloaded_archive_file(
    download_url: str,
    job_count: int = 4,
    retrieval_flags: int = 0,
) -> Generator[FileResults, None, None]:
    """
    Run the analysis on the given archive file after downloading it.

    :param download_url: The URL to download the archive from.
    :param job_count: The number of parallel jobs to use.
    :param retrieval_flags: Values to retrieve.
    :return: The requested results.
    """
    # Retrieving the correct suffixes is a bit tricky here, so we use some guessing as well.
    # This basically uses the trailing URL part (usually the filename itself) and forwards
    # guessing the suffixes to Python itself. Due to the way the suffixes are determined,
    # we probably get some part of the (irrelevant) filename as well as the suffix (due to
    # dotted package versions for example), but this should not hurt. We just have to make
    # sure that we do not really loose important suffix information as Python will not be
    # able to unpack this archive otherwise, for example because we supply `.gz` instead
    # of `.tar.gz` only.
    suffixes = Path(download_url.rsplit("/", maxsplit=1)[1]).suffixes
    suffix = "".join(suffixes)
    with NamedTemporaryFile(suffix=suffix) as downloaded_file:
        download_file(url=download_url, file_object=cast(BinaryIO, downloaded_file))
        yield from run_on_package_archive_file(
            archive_path=Path(downloaded_file.name),
            job_count=job_count,
            retrieval_flags=retrieval_flags,
        )


def run_on_downloaded_package_file(
    package_definition: str,
    index_url: str | None = None,
    job_count: int = 4,
    retrieval_flags: int = 0,
    prefer_sdist: bool = False,
) -> Generator[FileResults, None, None]:
    """
    Run the analysis for the given package definition.

    :param package_definition: The package definition to get the files for.
    :param index_url: The PyPI index URL to use. Uses the default one from the `.pypirc` file if unset.
    :param job_count: The number of parallel jobs to use.
    :param retrieval_flags: Values to retrieve.
    :param prefer_sdist: Download the source distribution instead of the wheel.
    :return: The requested results.
    """
    with TemporaryDirectory() as download_directory:
        download_package(
            package_definition=package_definition,
            download_directory=download_directory,
            index_url=index_url,
            prefer_sdist=prefer_sdist,
        )
        name = list(Path(download_directory).glob("*"))[0]
        yield from run_on_package_archive_file(
            archive_path=name.resolve(),
            job_count=job_count,
            retrieval_flags=retrieval_flags,
        )


def _check_that_exactly_one_value_is_set(values: list[Path | str | None]) -> bool:
    """
    Check that exactly one value does not evaluate to `False`.
    """
    filtered = list(filter(None, values))
    return len(filtered) == 1


def run(
    directory: Path | str | None = None,
    file_path: Path | str | None = None,
    archive_path: Path | str | None = None,
    package_definition: str | None = None,
    download_url: str | None = None,
    index_url: str | None = None,
    prefer_sdist: bool = False,
    job_count: int = 4,
    retrieve_copyrights: bool = False,
    retrieve_emails: bool = False,
    retrieve_file_info: bool = False,
    retrieve_urls: bool = False,
    retrieve_ldd_data: bool = False,
    retrieve_font_data: bool = False,
    retrieve_python_metadata: bool = False,
    retrieve_cargo_metadata: bool = False,
) -> list[FileResults]:
    """
    Run the analysis for the given input definition.

    The `directory`, `file_path`, `archive_path` and `package_definition` parameters are mutually exclusive,
    but exactly one has to be set.

    :param directory: The directory to run on.
    :param file_path: The file to run on.
    :param archive_path: The package archive to run on.
    :param package_definition: The package definition to run for.
    :param download_url: The package URL to download and run on.
    :param index_url: The PyPI index URL to use. Uses the default one from the `.pypirc` file if unset.
    :param prefer_sdist: For PyPI downloads, prefer/use the source distribution over/instead of the wheel.
    :param job_count: The number of parallel jobs to use.
    :param retrieve_copyrights: Whether to retrieve copyright information.
    :param retrieve_emails: Whether to retrieve e-mails.
    :param retrieve_file_info: Whether to retrieve file-specific information.
    :param retrieve_urls: Whether to retrieve URLs.
    :param retrieve_ldd_data: Whether to retrieve linking data for shared objects.
    :param retrieve_font_data: Whether to retrieve font data.
    :param retrieve_python_metadata: Whether to retrieve Python package metadata.
    :param retrieve_cargo_metadata: Whether to retrieve Cargo metadata.
    :return: The requested results.
    """
    # Remove the temporary directory of the main thread.
    atexit.register(scancode_tools.cleanup, scancode_config.scancode_temp_dir)

    assert _check_that_exactly_one_value_is_set(
        [directory, file_path, archive_path, package_definition, download_url]
    ), "Exactly one source is required."

    license_counts: dict[str | None, int] = defaultdict(int)
    retrieval_flags = RetrievalFlags.to_int(
        retrieve_copyrights=retrieve_copyrights,
        retrieve_emails=retrieve_emails,
        retrieve_file_info=retrieve_file_info,
        retrieve_urls=retrieve_urls,
        retrieve_ldd_data=retrieve_ldd_data,
        retrieve_font_data=retrieve_font_data,
        retrieve_python_metadata=bool(retrieve_python_metadata and package_definition),
        retrieve_cargo_metadata=retrieve_cargo_metadata,
    )

    # Run the analysis itself.
    if package_definition:
        results = list(
            run_on_downloaded_package_file(
                package_definition=package_definition,
                index_url=index_url,
                retrieval_flags=retrieval_flags,
                job_count=job_count,
                prefer_sdist=prefer_sdist,
            )
        )
    elif directory:
        results = list(
            run_on_directory(
                directory=str(directory),
                retrieval_flags=retrieval_flags,
                job_count=job_count,
            )
        )
    elif archive_path:
        results = list(
            run_on_package_archive_file(
                archive_path=Path(archive_path),
                retrieval_flags=retrieval_flags,
                job_count=job_count,
            )
        )
    elif download_url:
        results = list(
            run_on_downloaded_archive_file(
                download_url=download_url,
                retrieval_flags=retrieval_flags,
                job_count=job_count,
            )
        )
    elif file_path:
        results = [
            run_on_file(
                path=Path(file_path),
                short_path=str(file_path),
                retrieval_flags=retrieval_flags,
            )
         ]
    else:
        return []

    # Display the file-level results.
    max_path_length = max(len(result.short_path) for result in results)
    for result in results:
        if result.licenses is None:
            continue
        licenses = result.licenses
        scores = licenses.get_scores_of_detected_license_expression_spdx()
        print(
            f"{result.short_path:>{max_path_length}}",
            f"{licenses.detected_license_expression_spdx:>70}"
            if licenses.detected_license_expression_spdx
            else " " * 70,
            scores if scores else "",
        )
        license_counts[licenses.detected_license_expression_spdx] += 1

    # Display the license-level results.
    print()
    columns = shutil.get_terminal_size((80, 20)).columns
    print("=" * columns)
    print()
    count_length = max(math.log10(count) for count in license_counts.values())
    count_length = int(count_length) + 1
    for identifier in sorted(license_counts, key=str):
        print(
            f"{identifier!s:>70}", f"{license_counts[identifier]:>{count_length + 1}d}"
        )

    return results


atexit.register(scancode_tools.cleanup, scancode_config.scancode_temp_dir)



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Tools related to Cargo/Rust.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generator

import tomli

from license_tools.utils import download_utils, rendering_utils
from license_tools.utils.download_utils import Download

logger = logging.getLogger(__name__)
del logging


# https://doc.rust-lang.org/cargo/reference/manifest.html
_VERBOSE_NAMES = {
    "name": "Name",
    "version": "Version",
    "authors": "Authors",
    "description": "Description",
    "readme": "README",
    "homepage": "Homepage",
    "repository": "Repository",
    "license": "License",
    "license-file": "License File",
    "keywords": "Keywords",
    "categories": "Categories",
}


def read_toml(path: Path) -> dict[str, Any]:
    """
    Read the given TOML file.

    :param path: The file to read.
    :return: The parsed file content.
    """
    return tomli.loads(path.read_text())


def analyze_metadata(path: Path | str) -> dict[str, str | list[str]] | None:
    """
    Analyze the Rust package metadata for the given directory.

    :param path: The directory/file to analyze. Should either be a directory or `Cargo.toml` file.
    :return: The package metadata.
    """
    path = Path(path)
    if path.name != "Cargo.toml":
        if path.joinpath("Cargo.toml").exists():
            path = path / "Cargo.toml"
        elif len(list(path.glob("*"))) == 1:
            path = next(path.glob("*")) / "Cargo.toml"
        else:
            raise ValueError(f"No clear Cargo.toml in {path}.")
    manifest = read_toml(path)
    return manifest.get("package")


def check_metadata(path: Path | str) -> str:
    """
    Render the relevant details for the given package.

    :param path: The package path.
    :return: The rendered dictionary-like representation of the relevant fields.
    """
    metadata = analyze_metadata(path)
    if not metadata:
        return ""
    return rendering_utils.render_dictionary(
        dictionary=metadata, verbose_names_mapping=_VERBOSE_NAMES, multi_value_keys={"authors", "categories", "keywords"}
    )


@dataclass
class PackageVersion:
    """
    Container for holding a package version.
    """

    name: str
    """
    The package name.
    """

    version: str
    """
    The package version.
    """

    checksum: str
    """
    The package checksum.
    """

    def to_download(self) -> Download:
        """
        Generate the corresponding download URL.

        :return: The corresponding download.
        """
        return Download(
            url=f"https://crates.io/api/v1/crates/{self.name}/{self.version}/download",
            filename=f"{self.name}_{self.version}.crate",
            sha256=self.checksum
        )


def get_package_versions(lock_path: Path | str) -> Generator[PackageVersion, None, None]:
    """
    Get the packages from the given lock file.

    :param lock_path: The lock file to read.
    :return: The packages retrieved from lock file.
    """
    data = read_toml(Path(lock_path))
    for package in data["package"]:
        if package.get("source") != "registry+https://github.com/rust-lang/crates.io-index":
            logger.warning("Skipping %s", package)
            continue
        yield PackageVersion(name=package["name"], version=package["version"], checksum=package["checksum"])


def download_from_lock_file(lock_path: Path | str, target_directory: Path | str) -> None:
    """
    Download the packages from the given lock file.

    :param lock_path: The lock file to read.
    :param target_directory: The directory to write the packages to.
    """
    target_directory = Path(target_directory)
    if not target_directory.exists():
        target_directory.mkdir()

    downloads = [package.to_download() for package in get_package_versions(lock_path)]
    download_utils.download_one_file_per_second(downloads=downloads, directory=target_directory)



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import argparse
import logging
from typing import cast


def log_level_type(value: str | int) -> int:
    """
    Verify and convert a log level to an integer value.

    :param value: The value to validate/convert.
    :return: The corresponding integer value.
    """
    if isinstance(value, int):
        return value
    return cast(int, getattr(logging, value.upper()))


def configure_logging(level: int) -> None:
    """
    Configure the logging for the application.

    :param level: The log level to use.
    """
    logging.basicConfig(level=level)


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Run selected license tools. Will determine all licenses from the given source by default.",
    )

    source_group = parser.add_argument_group("Artifact source")
    source_group = source_group.add_mutually_exclusive_group(required=False)
    source_group.add_argument(
        "--directory", action="store", type=str, help="Directory to work on."
    )
    source_group.add_argument(
        "--file", action="store", type=str, help="File to work on."
    )
    source_group.add_argument(
        "--archive", action="store", type=str, help="Archive file to work on."
    )
    source_group.add_argument(
        "--package",
        action="store",
        type=str,
        help="Python package specification to use.",
    )
    source_group.add_argument(
        "--url", action="store", type=str, help="Download URL to use."
    )

    parser.add_argument(
        "--index-url",
        action="store",
        type=str,
        required=False,
        default="",
        help="PyPI index URL to use.",
    )
    parser.add_argument(
        "--prefer-sdist",
        action="store_true",
        default=False,
        required=False,
        help="Prefer/download the sdist over wheels on PyPI.",
    )

    parser.add_argument(
        "--jobs",
        action="store",
        type=int,
        required=False,
        default=4,
        help="Parallel jobs to use.",
    )

    parser.add_argument(
        "--retrieve-copyrights",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve copyrights.",
    )
    parser.add_argument(
        "--retrieve-emails",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve e-mails.",
    )
    parser.add_argument(
        "--retrieve-file-info",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve file information.",
    )
    parser.add_argument(
        "--retrieve-urls",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve URLs.",
    )
    parser.add_argument(
        "--retrieve-ldd-data",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve shared object linking data.",
    )
    parser.add_argument(
        "--retrieve-font-data",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve font metadata.",
    )
    parser.add_argument(
        "--retrieve-python-metadata",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve Python package metadata.",
    )
    parser.add_argument(
        "--retrieve-cargo-metadata",
        action="store_true",
        required=False,
        default=False,
        help="Retrieve Cargo/Rust package metadata.",
    )

    parser.add_argument(
        "--cargo-lock-download",
        action="store_true",
        required=False,
        default=False,
        help="Instead of analyzing the files, download the packages for a Cargo.lock file.",
    )
    parser.add_argument(
        "--cargo-lock",
        type=str,
        required=False,
        default=None,
        help="Path to the Cargo.lock file to use with `--cargo-lock-download`."
    )
    parser.add_argument(
        "--target-directory",
        type=str,
        required=False,
        default=None,
        help="Path to write the Cargo crate files to when using the `--cargo-lock-download` option."
    )

    parser.add_argument(
        "--log-level",
        type=log_level_type,
        required=False,
        default=logging.WARNING,
        help="Log level to use (name or integer). Defaults to ≥ warning."
    )

    arguments = parser.parse_args()
    configure_logging(level=arguments.log_level)

    if arguments.cargo_lock_download:
        from license_tools.tools import cargo_tools
        return cargo_tools.download_from_lock_file(
            lock_path=arguments.cargo_lock,
            target_directory=arguments.target_directory
        )

    from license_tools import retrieval
    retrieval.run(
        directory=arguments.directory,
        file_path=arguments.file,
        archive_path=arguments.archive,
        package_definition=arguments.package,
        download_url=arguments.url,
        index_url=arguments.index_url,
        prefer_sdist=arguments.prefer_sdist,
        job_count=arguments.jobs,
        retrieve_copyrights=arguments.retrieve_copyrights,
        retrieve_emails=arguments.retrieve_emails,
        retrieve_file_info=arguments.retrieve_file_info,
        retrieve_urls=arguments.retrieve_urls,
        retrieve_ldd_data=arguments.retrieve_ldd_data,
        retrieve_font_data=arguments.retrieve_font_data,
        retrieve_python_metadata=arguments.retrieve_python_metadata,
        retrieve_cargo_metadata=arguments.retrieve_cargo_metadata,
    )


if __name__ == "__main__":
    main()



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

from unittest import TestCase

from license_tools.utils import rendering_utils


class RenderDictionaryTestCase(TestCase):
    def test_render_dictionary(self) -> None:
        dictionary = {
            "key": "value",
            "pi": 3.141592653589793,
            "answer": 42,
            "key1": [1, 2, 3],
            "key2": (1, 2, 3),
            "key3": {3, 2, 1},
            "key4": {"foo": "bar"},
            "ignore": "me",
            "multi1": 1337,
            "multi2": [42, 43, 44],
        }
        mapping = {
            "key": "Key",
            "pi": "π",
            "answer": "Answer",
            "key1": "Key 1",
            "key2": "Key 2",
            "key3": "Key 3",
            "key4": "Key 4",
            "multi1": "Multiple 1",
            "multi2": "Multiple 2",
        }

        result = rendering_utils.render_dictionary(
            dictionary=dictionary,
            verbose_names_mapping=mapping,
            multi_value_keys={"key1", "key2", "key3", "multi1"},
        )
        self.assertEqual(
            """
       Key: value
         π: 3.141592653589793
    Answer: 42
     Key 1:
             * 1
             * 2
             * 3
     Key 2:
             * 1
             * 2
             * 3
     Key 3:
             * 1
             * 2
             * 3
     Key 4: {'foo': 'bar'}
Multiple 1: 1337
Multiple 2: [42, 43, 44]
"""[1:-1],
            result
        )



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock, TestCase

from license_tools.tools import linking_tools
from license_tools.tools.linking_tools import check_shared_objects
from tests import get_file
from tests.data import LICENSE_PATH, SETUP_PATH


def get_libc_path() -> Path:
    output = linking_tools.check_shared_objects(Path("/usr/bin/bc"))
    assert output is not None
    for line in output.splitlines(keepends=False):
        if "libc.so" in line:
            # 	libc.so.6 => /lib64/libc.so.6 (0x00007f281a209000)
            line = line.split("=>")[1]
            line = line.strip().split(" ")[0]
            return Path(line)
    raise ValueError("Could not determine libc.so location.")


class GetFileTypeTestCase(TestCase):
    def test_get_file_type(self) -> None:
        self.assertEqual(
            "Python script, ASCII text executable",
            linking_tools._get_file_type(SETUP_PATH)
        )
        self.assertEqual(
            "ASCII text",
            linking_tools._get_file_type(LICENSE_PATH)
        )

        with get_file("Carlito-Regular.ttf") as path:
            self.assertEqual(
                'TrueType Font data, 17 tables, 1st "GDEF", 15 names, Microsoft, language 0x409',
                linking_tools._get_file_type(path)
            )


class IsElfTestCase(TestCase):
    def test_is_elf(self) -> None:
        self.assertFalse(linking_tools.is_elf(SETUP_PATH))
        self.assertFalse(linking_tools.is_elf(LICENSE_PATH))

        with get_file("Carlito-Regular.ttf") as path:
            self.assertFalse(linking_tools.is_elf(path))

        self.assertTrue(linking_tools.is_elf(Path("/usr/bin/bc")))
        self.assertTrue(linking_tools.is_elf(get_libc_path().resolve()))  # Usually a symlink.


class GetElfTypeTestCase(TestCase):
    def test_get_elf_type(self) -> None:
        self.assertIsNone(linking_tools.get_elf_type(SETUP_PATH))
        self.assertIsNone(linking_tools.get_elf_type(LICENSE_PATH))

        with get_file("Carlito-Regular.ttf") as path:
            self.assertIsNone(linking_tools.get_elf_type(path))

        self.assertEqual(linking_tools.ELF_EXE, linking_tools.get_elf_type(Path("/usr/bin/bc")))
        self.assertEqual(linking_tools.ELF_SHARED, linking_tools.get_elf_type(get_libc_path().resolve()))  # Usually a symlink.


class CheckSharedObjectsTestCase(TestCase):
    def test_so(self) -> None:
        path = get_libc_path().resolve()
        with mock.patch(
            "subprocess.check_output", return_value=b"Test output\nAnother line\n"
        ) as subprocess_mock:
            result = check_shared_objects(path)
        self.assertEqual("Test output\nAnother line\n", result)
        subprocess_mock.assert_called_once_with(["ldd", path], stderr=subprocess.PIPE)

    def test_binary(self) -> None:
        path = Path("/usr/bin/bc")
        with mock.patch(
            "subprocess.check_output", return_value=b"Test output\nAnother line\n"
        ) as subprocess_mock:
            result = check_shared_objects(path)
        self.assertEqual("Test output\nAnother line\n", result)
        subprocess_mock.assert_called_once_with(["ldd", path], stderr=subprocess.PIPE)

    def test_python(self) -> None:
        path = Path("/tmp/libdummy.py")
        with mock.patch(
            "subprocess.check_output", return_value=b"Test output\nAnother line\n"
        ) as subprocess_mock:
            result = check_shared_objects(path)
        self.assertIsNone(result)
        subprocess_mock.assert_not_called()

    def test_symlink(self) -> None:
        with TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            target = directory.joinpath("libdummy.so.42.0.0")
            target.write_bytes(b"abc")
            source = directory.joinpath("libdummy.so")
            source.symlink_to(target=target)

            with mock.patch(
                "subprocess.check_output", return_value=b"Test output\nAnother line\n"
            ) as subprocess_mock, mock.patch.object(
                linking_tools.logger, "warning"
            ) as warning_mock, mock.patch.object(
                linking_tools, "is_elf", return_value=True
            ) as elf_mock:
                result = check_shared_objects(source)
            self.assertIsNone(result)
            subprocess_mock.assert_not_called()
            warning_mock.assert_called_once_with(
                "Ignoring symlink %s to %s for shared object analysis.", source, target
            )
            elf_mock.assert_called_once_with(source)

            with mock.patch(
                "subprocess.check_output", return_value=b"Test output\nAnother line\n"
            ) as subprocess_mock, mock.patch.object(
                linking_tools, "is_elf", return_value=True
            ) as elf_mock:
                result = check_shared_objects(source.resolve())
            self.assertEqual("Test output\nAnother line\n", result)
            subprocess_mock.assert_called_once_with(
                ["ldd", target], stderr=subprocess.PIPE
            )
            elf_mock.assert_called_once_with(source.resolve())



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Tools related to fonts.
"""

from __future__ import annotations

import datetime
from collections import OrderedDict
from pathlib import Path
from typing import Any, Union

from fontTools import ttx  # type: ignore[import-untyped]
from fontTools.misc import timeTools  # type: ignore[import-untyped]
from fontTools.ttLib import TTFont  # type: ignore[import-untyped]
from fontTools.ttLib.tables._h_e_a_d import table__h_e_a_d as HeadTable  # type: ignore[import-untyped]  # noqa: N812
from fontTools.ttLib.tables._n_a_m_e import table__n_a_m_e as NameTable  # type: ignore[import-untyped]  # noqa: N812


# https://learn.microsoft.com/en-us/typography/opentype/spec/name#name-ids
_TTF_NAME_IDS: list[str] = [
    "Copyright notice",
    "Font family name",
    "Font subfamily name",
    "Unique font identifier",
    "Full font name",
    "Version string",
    "PostScript name",
    "Trademark",
    "Manufacturer",
    "Designer",
    "Description",
    "URL Vendor",
    "URL Designer",
    "License Description",
    "License Info URL",
    "Reserved",
    "Typographic Family name",
    "Typographic Subfamily name",
    "Compatible Full (Mac only)",
    "Sample text",
    "PostScript CID findfont name",
    "WWS Family Name",
    "WWS Subfamily Name",
    "Light Background Palette",
    "Dark Background Palette",
    "Variations PostScript Name Prefix",
]


# https://learn.microsoft.com/en-us/typography/opentype/spec/head
def convert_head_flags(value: int) -> str:
    """
    Convert the given head flags value to a human-readable string.

    :param value: The flags value.
    :return: The corresponding human-readable interpretation.
    """
    result = []
    verbose = [
        "Baseline for font at y=0",
        "Left sidebearing point at x=0",
        "Instructions may depend on point size",
        "Force ppem to integer values",
        "Instructions may alter advance width",
        "",
        "",
        "",
        "",
        "",
        "",
        "Lossless font data",
        "Font converted",
        "Font optimized for ClearType",
        "Last Resort font",
        "",
    ]
    for index, verbose_ in enumerate(verbose):
        if value & (1 << index):
            result.append(verbose_)
    return "; ".join(filter(None, result)) if result else "%"


def convert_timestamp_to_datetime(value: int) -> datetime.datetime:
    """
    Convert the given font timestamp to a datetime object.

    :param value: The font timestamp since the font epoch in 1904.
    :return: The regular datetime object.
    """
    return datetime.datetime.fromtimestamp(
        max(0, value) + timeTools.epoch_diff, tz=datetime.timezone.utc
    )


def convert_timestamp_to_string(value: int) -> str:
    """
    Convert the given font timestamp to a string.

    :param value: The font timestamp since the font epoch in 1904.
    :return: The corresponding string representation.
    """
    return convert_timestamp_to_datetime(value).strftime("%Y-%m-%d %H:%M:%S")


def convert_mac_style(value: int) -> str:
    """
    Convert the mac style to a human-readable representation.

    :param value: The style flags.
    :return: The corresponding string representation.
    """
    result = []
    verbose = [
        "Bold",
        "Italic",
        "Underline",
        "Outline",
        "Shadow",
        "Condensed",
        "Extended",
    ]
    for index, verbose_ in enumerate(verbose):
        if value & (1 << index):
            result.append(verbose_)
    return ", ".join(result) if result else "%"


def convert_font_direction_hint(value: int) -> str:
    """
    Convert the given font direction hint to a human-readable representation.

    :param value: The hint value.
    :return: The human-readable representation.
    """
    return {
        0: "Fully mixed directional glyphs",
        1: "Only strongly left to right",
        2: "Strongly left to right, but also contains neutrals",
        -1: "Only strongly right to left",
        -2: "Strongly right to left, but also contains neutrals",
    }[value]


def convert_loc_format(value: int) -> str:
    """
    Convert the given loc value to a human-readable representation.

    :param value: The loc value.
    :return: The human-readable representation.
    """
    return {
        0: "Short offsets (Offset16)",
        1: "Long offsets (Offset32)",
    }[value]


def identity(value: Any) -> Any:
    """
    Utility function to return the value itself.

    :param value: The value to return.
    :return: The input value.
    """
    return value


_TTF_HEAD_IDS = {
    "tableVersion": ("Font Table Version", identity),
    "fontRevision": ("Font Revision", identity),
    "checkSumAdjustment": ("Checksum", identity),
    "magicNumber": ("Magic number", identity),
    "flags": ("Flags", convert_head_flags),
    "unitsPerEm": ("Units per em", identity),
    "created": ("Created", convert_timestamp_to_string),
    "modified": ("Modified", convert_timestamp_to_string),
    "xMin": ("xMin", identity),
    "yMin": ("yMin", identity),
    "xMax": ("xMax", identity),
    "yMax": ("yMax", identity),
    "macStyle": ("Mac Style", convert_mac_style),
    "lowestRecPPEM": ("Smallest readable size in pixels", identity),
    "fontDirectionHint": ("Font direction hint", convert_font_direction_hint),
    "indexToLocFormat": ("Index to Loc format", convert_loc_format),
    "glyphDataFormat": ("Glyph Data Format", identity),
}

KNOWN_FONT_EXTENSIONS = {".otf", ".ttf", ".woff", ".woff2"}

FONT_VALUE_TYPE = Union[str, int, datetime.datetime]


def handle_ttfont_head(head: HeadTable) -> OrderedDict[str, FONT_VALUE_TYPE]:
    """
    Handle the head section of a TTFont instance.

    :param head: The section to handle.
    :return: The parsed/human-readable mapping.
    """
    result = OrderedDict()
    for key, value in head.__dict__.items():
        if key not in _TTF_HEAD_IDS:
            continue
        identifier, value_parser = _TTF_HEAD_IDS[key]
        parsed_value = value_parser(value)
        result[identifier] = parsed_value
    return result


def handle_ttfont_names(names: NameTable) -> OrderedDict[str, FONT_VALUE_TYPE]:
    """
    Handle the names section of a TTFont instance.

    :param names: The names to handle.
    :return: The parsed/human-readable mapping.
    """
    result = OrderedDict()
    for index, verbose_name in enumerate(_TTF_NAME_IDS):
        name = names.getDebugName(index)
        if name:
            result[verbose_name] = name
    return result


def analyze_font(path: Path) -> dict[str, None | dict[str, FONT_VALUE_TYPE]] | None:
    """
    Analyze the given font and provide a human-readable mapping for the head and name
    sections.

    :param path: The font path.
    :return: `None` if this is no known font type, the parsed results grouped by
             section name otherwise.
    """
    if path.suffix not in KNOWN_FONT_EXTENSIONS:
        return None

    result: dict[str, None | dict[str, FONT_VALUE_TYPE]] = {"head": None, "name": None}
    with TTFont(file=path) as font:
        for key in font.keys():
            if key not in {"head", "name"}:
                continue
            value = font.get(key)
            if key == "head":
                result[key] = handle_ttfont_head(value)
            elif key == "name":
                result[key] = handle_ttfont_names(value)
    return result


def check_font(path: Path) -> str | None:
    """
    Render the relevant details for the given font.

    :param path: The font path.
    :return: `None` if no results could be determined, otherwise the rendered
             dictionary-like representation of the names section which usually
             holds the copyright data.
    """
    font_data = analyze_font(path)
    if not font_data:
        return None
    names = font_data["name"]
    if not names:
        return None
    maximum_length = max(map(len, names.keys()))
    rendered = "\n".join(
        f"{key:>{maximum_length}}: {value}" for key, value in names.items()
    )
    return rendered


def dump_to_ttx(source_path: Path, target_path: Path) -> Path:
    """
    Utility function to convert the given font to its XML representation.

    :param source_path: The font to read.
    :param target_path: The path to write to.
    :return: The target path. The same as the `target_path` if its suffix is `.ttx`,
             otherwise the `target_path` with the suffix added.
    """
    if target_path.suffix != ".ttx":
        target_path = target_path.parent / (target_path.name + ".ttx")
    ttx.ttDump(
        input=source_path,
        output=target_path,
        options=ttx.Options(rawOptions=[], numFiles=1),
    )
    return target_path



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Convenience interface for the ScanCode toolkit project using some predefined
configuration and returning `dataclass` instances instead of dictionaries.
"""

from __future__ import annotations

import atexit
import datetime
from dataclasses import dataclass, field as dataclass_field
from pathlib import Path
from typing import Literal

import scancode_config  # type: ignore[import-untyped]
from commoncode import fileutils  # type: ignore[import-untyped]
from packagedcode.rpm import RpmArchiveHandler  # type: ignore[import-untyped]
from scancode import api  # type: ignore[import-untyped]


@dataclass
class Author:
    """
    Matching information about an author.
    """

    author: str
    """
    The author name.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """


@dataclass
class Holder:
    """
    Matching information about a copyright holder.
    """

    holder: str
    """
    The copyright holder.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """


@dataclass
class Copyright:
    """
    Matching information about copyrights.
    """

    copyright: str
    """
    The copyright statement.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """


@dataclass
class Copyrights:
    """
    Copyright-specific results.
    """

    copyrights: list[Copyright] = dataclass_field(default_factory=list)
    """
    The detected copyright statements.
    """

    holders: list[Holder] = dataclass_field(default_factory=list)
    """
    The detected copyright holders.
    """

    authors: list[Author] = dataclass_field(default_factory=list)
    """
    The detected authors.
    """

    def __post_init__(self) -> None:
        self.copyrights = [Copyright(**x) if not isinstance(x, Copyright) else x for x in self.copyrights]  # type: ignore[arg-type]
        self.holders = [Holder(**x) if not isinstance(x, Holder) else x for x in self.holders]  # type: ignore[arg-type]
        self.authors = [Author(**x) if not isinstance(x, Author) else x for x in self.authors]  # type: ignore[arg-type]


@dataclass
class Email:
    """
    Matching information about an e-mail.
    """

    email: str
    """
    The e-mail address.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """


@dataclass
class Emails:
    """
    E-mail-specific results.
    """

    emails: list[Email] = dataclass_field(default_factory=list)
    """
    The detected e-mail addresses.
    """

    def __post_init__(self) -> None:
        self.emails = [Email(**x) if not isinstance(x, Email) else x for x in self.emails]  # type: ignore[arg-type]


@dataclass
class Url:
    """
    Matching information about an URL.
    """

    url: str
    """
    The URL.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """


@dataclass
class Urls:
    """
    URL-specific results.
    """

    urls: list[Url] = dataclass_field(default_factory=list)
    """
    The detected URLs.
    """

    def __post_init__(self) -> None:
        self.urls = [Url(**x) if not isinstance(x, Url) else x for x in self.urls]  # type: ignore[arg-type]


@dataclass
class FileInfo:
    """
    File-specific results.
    """

    date: datetime.date
    """
    The modification date.
    """

    size: int
    """
    The file size in bytes.
    """

    sha1: str | None
    """
    The SHA1 hash.
    """

    md5: str | None
    """
    The MD5 hash.
    """

    sha256: str | None
    """
    The SHA256 hash.
    """

    mime_type: str
    """
    The detected mime type.
    """

    file_type: str
    """
    The detected file type.
    """

    programming_language: str | None
    """
    The detected programming language.
    """

    is_binary: bool
    """
    Whether this file is a binary one.
    """

    is_text: bool
    """
    Whether this file is a plaintext one.
    """

    is_archive: bool
    """
    Whether this file is an archive.
    """

    is_media: bool
    """
    Whether this is some media file.
    """

    is_source: bool
    """
    Whether this is some source code file.
    """

    is_script: bool
    """
    Whether this file is some script.
    """

    def __post_init__(self) -> None:
        if isinstance(self.date, str):
            self.date = datetime.datetime.strptime(self.date, "%Y-%m-%d").date()


@dataclass
class LicenseMatch:
    """
    Matching information about a license.
    """

    score: float
    """
    The matching score.
    """

    start_line: int
    """
    The corresponding start line.
    """

    end_line: int
    """
    The corresponding end line.
    """

    matched_length: int
    """
    The length of the match in bytes/characters.
    """

    match_coverage: float
    """
    How much of the rule text is part of the match?
    """

    matcher: str
    """
    The corresponding matcher type.
    """

    license_expression: str
    """
    The detected license expression.
    """

    spdx_license_expression: str
    """
    The detected license expression in SPDX format.
    """

    rule_identifier: str
    """
    The corresponding matcher rule.
    """

    rule_relevance: int
    """
    The relevance of the corresponding rule.
    """

    rule_url: str | None
    """
    Upstream link for this rule.
    """

    from_file: str | None
    """
    Unused/unclear.
    """

    matched_text: str | None = None
    """
    The text of the match.
    """


@dataclass
class LicenseClue(LicenseMatch):
    """
    Matching information about a license. Compared to regular detections/matches, these are rather
    considered clues and not perfect detections.

    Currently the same as :class:`~LicenseMatch`.
    """

    pass


@dataclass
class LicenseDetection:
    """
    Information on a specific detected license.
    """

    license_expression: str
    """
    The detected license expression.
    """

    license_expression_spdx: str
    """
    The detected license expression in SPDX format.
    """

    identifier: str
    """
    An unique ID for this detection.
    """

    matches: list[LicenseMatch] = dataclass_field(default_factory=list)
    """
    The corresponding detailed match data.
    """

    def __post_init__(self) -> None:
        self.matches = [LicenseMatch(**x) if not isinstance(x, LicenseMatch) else x for x in self.matches]  # type: ignore[arg-type]


@dataclass
class Licenses:
    """
    Information on all detected licenses.
    """

    detected_license_expression: str | None = None
    """
    The detected license expression.
    """

    detected_license_expression_spdx: str | None = None
    """
    The detected license expression in SPDX format.
    """

    percentage_of_license_text: float = 0.0
    """
    How much of the file content is part of the license text?
    """

    license_detections: list[LicenseDetection] = dataclass_field(default_factory=list)
    """
    The corresponding license detections.
    """

    license_clues: list[LicenseClue] = dataclass_field(default_factory=list)
    """
    The corresponding license clues.
    """

    def __post_init__(self) -> None:
        self.license_detections = [
            LicenseDetection(**x) if not isinstance(x, LicenseDetection) else x for x in self.license_detections  # type: ignore[arg-type]
        ]
        self.license_clues = [
            LicenseClue(**x) if not isinstance(x, LicenseClue) else x for x in self.license_clues  # type: ignore[arg-type]
        ]

    def get_scores_of_detected_license_expression_spdx(self) -> list[float]:
        """
        Attempt to resolve the scores for the detected license expression.

        :return: The corresponding scores if they could be resolved.
        """
        scores = []
        for detection in self.license_detections:
            if (
                detection.license_expression == self.detected_license_expression
                or detection.license_expression_spdx
                == self.detected_license_expression_spdx
            ):
                for match in detection.matches:
                    scores.append(match.score)
                return scores
        return scores


@dataclass
class FileResults:
    """
    Container for all available file-level results.
    """

    # Reference to the analyzed file.
    path: Path
    """
    Full path of the analyzed file.
    """

    short_path: str
    """
    Short path of the analyzed file, with the common prefix removed.
    """

    # Configuration values to determine which information to retrieve.
    retrieve_copyrights: bool = False
    """
    Configuration option: Whether to retrieve copyright information.
    """

    retrieve_emails: bool = False
    """
    Configuration option: Whether to retrieve e-mail addresses.
    """

    retrieve_urls: bool = False
    """
    Configuration option: Whether to retrieve URLs.
    """

    retrieve_licenses: bool = False
    """
    Configuration option: Whether to retrieve license information.
    """

    retrieve_file_info: bool = False
    """
    Configuration option: Whether to retrieve file information.
    """

    # Analysis results.
    copyrights: Copyrights | None = None
    """
    The detected copyrights.
    """

    emails: Emails | None = None
    """
    The detected e-mail addresses.
    """

    urls: Urls | None = None
    """
    The detected URLs.
    """

    licenses: Licenses | None = None
    """
    The detected licenses.
    """

    file_info: FileInfo | None = None
    """
    The retrieved file information.
    """

    def __post_init__(self) -> None:
        path_str = str(self.path)
        if self.retrieve_copyrights:
            self.copyrights = Copyrights(**api.get_copyrights(path_str))
        if self.retrieve_emails:
            self.emails = Emails(**api.get_emails(path_str))
        if self.retrieve_urls:
            self.urls = Urls(**api.get_urls(path_str))
        if self.retrieve_licenses:
            self.licenses = Licenses(**api.get_licenses(path_str))
        if self.retrieve_file_info:
            self.file_info = FileInfo(**api.get_file_info(path_str))


@dataclass
class Party:
    """
    A party related to a package.
    """

    PARTY_TYPES = Literal[None, "person", "project", "organization"]
    """
    Available party types.
    """

    type: PARTY_TYPES = None
    """
    The party type.
    """

    role: str | None = None
    """
    The associated role.
    """

    name: str | None = None
    """
    The corresponding name.
    """

    email: str | None = None
    """
    The corresponding e-mail address.
    """

    url: str | None = None
    """
    The corresponding URL/webpage.
    """


@dataclass
class PackageResults:
    """
    Container for package-specific data, based upon ``packagedcode.models.PackageData``.
    """

    api_data_url: str | None = None
    """
    API URL to obtain structured data for this package.
    """

    bug_tracking_url: str | None = None
    """
    URL to the issue or bug tracker.
    """

    code_view_url: str | None = None
    """
    An URL where th ecode can be browsed online.
    """

    copyright: str | None = None
    """
    Copyright statements for this package.
    """

    datasource_id: str | None = None
    """
    Data source identifier for the source of the package data.
    """

    declared_license_expression: str | None = None
    """
    License expression derived from metadata.
    """

    declared_license_expression_spdx: str | None = None
    """
    License expression derived from metadata in SPDX format.
    """

    # dependencies: list[DependentPackage] = dataclass_field(default_factory=list)
    description: str | None = None
    """
    Description for this package.
    """

    download_url: str | None = None
    """
    A direct download URL.
    """

    # extra_data: dict[str, Any] = dataclass_field(default_factory=dict)
    extracted_license_statement: str | None = None
    """
    License statement extracted from the metadata.
    """

    # file_references: list[FileReference] = dataclass_field(default_factory=list)
    holder: str | None = None
    """
    Copyright holders for this package.
    """

    homepage_url: str | None = None
    """
    URL to the homepage.
    """

    keywords: list[str] = dataclass_field(default_factory=list)
    """
    Associated keywords.
    """

    license_detections: list[LicenseDetection] = dataclass_field(default_factory=list)
    """
    Detected licenses.
    """

    md5: str | None = None
    """
    MD5 checksum.
    """

    name: str | None = None
    """
    Package name.
    """

    namespace: str | None = None
    """
    Package namespace.
    """

    notice_text: str | None = None
    """
    Corresponding notice text.
    """

    other_license_detections: list[LicenseDetection] = dataclass_field(
        default_factory=list
    )
    """
    Additional license detections.
    """

    other_license_expression: str | None = None
    """
    Another/a secondary license derived from the metadata.
    """

    other_license_expression_spdx: str | None = None
    """
    Another/a secondary license derived from the metadata in SPDX format.
    """

    parties: list[Party] = dataclass_field(default_factory=list)
    """
    Parties such as a person, project or organization.
    """

    primary_language: str | None = None
    """
    Primary programming language.
    """

    purl: str | None = None
    """
    Corresponding PURL identifier.
    """

    qualifiers: dict[str, str] = dataclass_field(default_factory=dict)
    """
    Mapping of package qualifiers.
    """

    release_date: datetime.date | None = None
    """
    Release date of the package.
    """

    repository_download_url: str | None = None
    """
    Download URL for this package in its package repository.
    """

    repository_homepage_url: str | None = None
    """
    URL to the page for this package in its package repository.
    """

    sha1: str | None = None
    """
    SHA1 checksum.
    """

    sha256: str | None = None
    """
    SHA256 checksum.
    """

    sha512: str | None = None
    """
    SHA512 checksum.
    """

    size: int | None = None
    """
    Size of the package download in bytes.
    """

    source_packages: list[str] | None = None
    """
    PURLs for related source packages.
    """

    subpath: str | None = None
    """
    Subpath inside a package, relative to its root.
    """

    type: str | None = None
    """
    Short code for package type.
    """

    vcs_url: str | None = None
    """
    An URL to the VCS repository in SPDX format.
    """

    version: str | None = None
    """
    Package version.
    """

    is_virtual: bool = False
    """
    Indicate virtual packages.
    """

    is_private: bool = False
    """
    Indicate private packages.
    """

    def __post_init__(self) -> None:
        self.license_detections = [
            LicenseDetection(**x) if not isinstance(x, LicenseDetection) else x for x in self.license_detections  # type: ignore[arg-type]
        ]
        self.other_license_detections = [
            LicenseDetection(**x) if not isinstance(x, LicenseDetection) else x for x in self.other_license_detections  # type: ignore[arg-type]
        ]
        self.parties = [
            Party(**x) if not isinstance(x, Party) else x for x in self.parties  # type: ignore[arg-type]
        ]

    @classmethod
    def from_rpm(cls, path: Path) -> "PackageResults":
        """
        Get the results for the given RPM path.

        :param path: The RPM path to run on.
        :return: The corresponding results.
        """
        # Drop some keys which we do not handle for now.
        data = next(RpmArchiveHandler.parse(path)).to_dict()
        data.pop("dependencies", None)
        data.pop("extra_data", None)
        data.pop("file_references", None)
        return cls(**data)


def cleanup(directory: Path | str) -> None:
    """
    Remove the given directory.
    """
    fileutils.delete(directory)


atexit.register(cleanup, scancode_config.scancode_temp_dir)



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import datetime
from pathlib import Path
from tempfile import mkdtemp, NamedTemporaryFile
from typing import cast, Dict
from unittest import TestCase

from license_tools.retrieval import RetrievalFlags
from license_tools.tools import scancode_tools
from license_tools.tools.scancode_tools import (
    Copyright,
    Copyrights,
    Emails,
    FileInfo,
    FileResults,
    Holder,
    LicenseDetection, LicenseMatch, Licenses,
    PackageResults, Party, Url,
    Urls,
)
from tests import get_from_url
from tests.data import LIBAIO1__0_3_109_1_25__RPM, LICENSE_PATH, SETUP_PATH, SETUP_PY_LICENSES


class LicensesTestCase(TestCase):
    def test_get_scores_of_detected_license_expression_spdx(self) -> None:
        result = SETUP_PY_LICENSES.get_scores_of_detected_license_expression_spdx()
        self.assertEqual([], result)

        file_results = FileResults(
            path=LICENSE_PATH, short_path="LICENSE.txt", retrieve_licenses=True
        )
        licenses = cast(Licenses, file_results.licenses)
        result = licenses.get_scores_of_detected_license_expression_spdx()
        self.assertEqual([100.0], result)


class FileResultsTestCase(TestCase):
    def assert_not_requested(
        self, result: FileResults, fields: list[str], invert: bool = False
    ) -> None:
        method = self.assertIsNotNone if invert else self.assertIsNone
        for field in fields:
            with self.subTest(field=field):
                method(getattr(result, field))

    def test_full(self) -> None:
        flags = cast(Dict[str, bool], RetrievalFlags.all(as_kwargs=True))
        del flags["retrieve_ldd_data"]
        del flags["retrieve_font_data"]
        del flags["retrieve_python_metadata"]
        del flags["retrieve_cargo_metadata"]
        self.assertEqual(4, len(flags))
        result = FileResults(
            path=SETUP_PATH, short_path="setup.py", retrieve_licenses=True,
            retrieve_copyrights=flags["retrieve_copyrights"],
            retrieve_emails=flags["retrieve_emails"],
            retrieve_urls=flags["retrieve_urls"],
            retrieve_file_info=flags["retrieve_file_info"],
        )
        self.assertEqual(SETUP_PATH, result.path)
        self.assertEqual("setup.py", result.short_path)
        self.assert_not_requested(
            result=result,
            fields=["copyrights", "emails", "urls", "licenses", "file_info"],
            invert=True,
        )

    def test_none(self) -> None:
        result = FileResults(path=Path("/tmp/dummy.py"), short_path="dummy.py")
        self.assertEqual(Path("/tmp/dummy.py"), result.path)
        self.assertEqual("dummy.py", result.short_path)
        self.assert_not_requested(
            result=result,
            fields=["copyrights", "emails", "urls", "licenses", "file_info"],
        )

    def test_retrieve_copyrights(self) -> None:
        result = scancode_tools.FileResults(
            path=SETUP_PATH, short_path="setup.py", retrieve_copyrights=True
        )
        self.assertEqual(SETUP_PATH, result.path)
        self.assertEqual("setup.py", result.short_path)
        self.assert_not_requested(
            result=result, fields=["emails", "urls", "licenses", "file_info"]
        )
        expected = Copyrights(
            copyrights=[
                Copyright(
                    copyright="Copyright (c) stefan6419846", start_line=1, end_line=1
                )
            ],
            holders=[Holder(holder="stefan6419846", start_line=1, end_line=1)],
            authors=[],
        )
        self.assertEqual(expected, result.copyrights)

    def test_retrieve_emails(self) -> None:
        result = scancode_tools.FileResults(
            path=SETUP_PATH, short_path="setup.py", retrieve_emails=True
        )
        self.assertEqual(SETUP_PATH, result.path)
        self.assertEqual("setup.py", result.short_path)
        self.assert_not_requested(
            result=result, fields=["copyrights", "urls", "licenses", "file_info"]
        )
        expected = Emails(emails=[])
        self.assertEqual(expected, result.emails)

    def test_retrieve_urls(self) -> None:
        result = scancode_tools.FileResults(
            path=SETUP_PATH, short_path="setup.py", retrieve_urls=True
        )
        self.assertEqual(SETUP_PATH, result.path)
        self.assertEqual("setup.py", result.short_path)
        self.assert_not_requested(
            result=result, fields=["copyrights", "emails", "licenses", "file_info"]
        )
        expected = Urls(
            urls=[
                Url(
                    url="http://www.apache.org/licenses/LICENSE-2.0",
                    start_line=3,
                    end_line=3,
                ),
                Url(
                    url="https://github.com/stefan6419846/license_tools",
                    start_line=21,
                    end_line=21,
                ),
            ]
        )
        self.assertEqual(expected, result.urls)

    def test_retrieve_licenses(self) -> None:
        self.maxDiff = None
        result = scancode_tools.FileResults(
            path=SETUP_PATH, short_path="setup.py", retrieve_licenses=True
        )
        self.assertEqual(SETUP_PATH, result.path)
        self.assertEqual("setup.py", result.short_path)
        self.assert_not_requested(
            result=result, fields=["copyrights", "emails", "urls", "file_info"]
        )
        self.assertEqual(SETUP_PY_LICENSES, result.licenses)

    def test_retrieve_file_info(self) -> None:
        with NamedTemporaryFile(suffix=".py") as file_object:
            path = Path(file_object.name)
            path.write_text('print("Hello World!")\n')
            result = scancode_tools.FileResults(
                path=path, short_path="test.py", retrieve_file_info=True
            )

        self.assertEqual(path, result.path)
        self.assertEqual("test.py", result.short_path)
        self.assert_not_requested(
            result=result, fields=["copyrights", "emails", "urls", "licenses"]
        )
        expected = FileInfo(
            date=datetime.date.today(),
            size=22,
            sha1="e343a35cf2fa04782749dab102d45129cdb0b644",
            md5="a97c0affb458a65d8682bf0a48f36e63",
            sha256="c63bf759e5502fc9f4ad863b883423a2d75992aeaebee6a713eb81fe3f714a4b",
            mime_type="text/plain",
            file_type="ASCII text",
            programming_language="Python",
            is_binary=False,
            is_text=True,
            is_archive=False,
            is_media=False,
            is_source=True,
            is_script=False,
        )
        self.assertEqual(expected, result.file_info)


class PackageResultsTestCase(TestCase):
    def test_rpm(self) -> None:
        with get_from_url(LIBAIO1__0_3_109_1_25__RPM) as rpm_path:
            results = PackageResults.from_rpm(rpm_path)
        self.assertEqual(
            PackageResults(
                api_data_url=None, bug_tracking_url=None, code_view_url=None, copyright=None, datasource_id='rpm_archive',
                declared_license_expression='lgpl-2.1-plus', declared_license_expression_spdx='LGPL-2.1-or-later',
                description=(
                    'Linux-Native Asynchronous I/O Access Library\nThe Linux-native asynchronous I/O facility ("async I/O", or "aio") has\n'
                    'a richer API and capability set than the simple POSIX async I/O\nfacility. This library provides the Linux-native API for async I/O. '
                    'The\nPOSIX async I/O facility requires this library to provide\nkernel-accelerated async I/O capabilities, as do applications that\n'
                    'require the Linux-native async I/O API.'
                ),
                download_url=None, extracted_license_statement='LGPL-2.1+', holder=None,
                homepage_url='http://kernel.org/pub/linux/libs/aio/', keywords=[],
                license_detections=[
                    LicenseDetection(
                        license_expression='lgpl-2.1-plus', license_expression_spdx='LGPL-2.1-or-later',
                        identifier='lgpl_2_1_plus-d0bddf88-13b4-4982-bdf5-ce866a0e8394',
                        matches=[
                            LicenseMatch(
                                score=100.0, start_line=1, end_line=1, matched_length=3, match_coverage=100.0, matcher='1-hash',
                                license_expression='lgpl-2.1-plus', rule_identifier='lgpl-2.1-plus_64.RULE', rule_relevance=100,
                                rule_url='https://github.com/nexB/scancode-toolkit/tree/develop/src/licensedcode/data/rules/lgpl-2.1-plus_64.RULE',
                                matched_text='LGPL-2.1+',
                                spdx_license_expression='LGPL-2.1-or-later', from_file=None,
                            )
                        ]
                    )
                ],
                md5=None, name='libaio1', namespace=None, notice_text=None, other_license_detections=[], other_license_expression=None,
                other_license_expression_spdx=None,
                parties=[
                    Party(type=None, role='distributor', name='SUSE Linux Enterprise 15', email=None, url=None),
                    Party(type=None, role='vendor', name='SUSE LLC <https://www.suse.com/>', email=None, url=None)
                ],
                primary_language=None, purl='pkg:rpm/libaio1@0.3.109-1.25', qualifiers={}, release_date=None,
                repository_download_url=None, repository_homepage_url=None, sha1=None, sha256=None, sha512=None, size=None,
                source_packages=['pkg:rpm/libaio@0.3.109-1.25?arch=src'], subpath=None, type='rpm', vcs_url=None, version='0.3.109-1.25'
            ),
            results
        )


class CleanupTestCase(TestCase):
    def test_cleanup(self) -> None:
        directory_string = mkdtemp()
        directory = Path(directory_string)
        self.assertTrue(directory.is_dir())

        # 1) Existing directory.
        scancode_tools.cleanup(directory)
        self.assertFalse(directory.is_dir())

        # 2) Missing directory.
        scancode_tools.cleanup(directory)
        self.assertFalse(directory.is_dir())



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import datetime
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest import mock, TestCase

import requests

from license_tools.utils import download_utils
from license_tools.utils.download_utils import ChecksumError, Download, DownloadError
from license_tools.utils.path_utils import get_files_from_directory


class DownloadTestCase(TestCase):
    def test_verify_checksum(self) -> None:
        # No checksum.
        Download(url="http://localhost", filename="dummy").verify_checksum(b"Hello World!\n")

        # Correct sha256 checksum.
        Download(
            url="http://localhost", filename="dummy", sha256="03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340"
        ).verify_checksum(b"Hello World!\n")

        # Incorrect sha256 checksum.
        with self.assertRaisesRegex(
                expected_exception=ChecksumError,
                expected_regex=r"^Checksum mismatch: Got 03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340, expected INVALID!$"
        ):
            Download(url="http://localhost", filename="dummy", sha256="INVALID").verify_checksum(b"Hello World!\n")


class GetSessionTestCase(TestCase):
    def test_get_session(self) -> None:
        session = download_utils.get_session()
        self.assertIsInstance(session, requests.Session)
        self.assertIn("https://github.com/stefan6419846/license_tools version ", session.headers["User-Agent"])


class DownloadFileTestCase(TestCase):
    def test_not_okay(self) -> None:
        session = download_utils.get_session()
        response = requests.Response()
        response.status_code = 404
        file_object = BytesIO()
        with mock.patch.object(download_utils, "get_session", return_value=session), \
                mock.patch.object(session, "get", return_value=response):
            with self.assertRaisesRegex(
                    expected_exception=DownloadError,
                    expected_regex=r"^Download not okay\? http://localhost <Response \[404\]>$"
            ):
                download_utils.download_file(
                    url="http://localhost",
                    file_object=file_object,
                )
        self.assertEqual(b"", file_object.getvalue())

    def test_valid(self) -> None:
        session = download_utils.get_session()
        response = requests.Response()
        response.status_code = 200
        response._content = b"Hello World!\n"
        file_object = BytesIO()

        with mock.patch.object(download_utils, "get_session", return_value=session), \
                mock.patch.object(session, "get", return_value=response):
            download_utils.download_file(
                url="http://localhost",
                file_object=file_object,
            )
        self.assertEqual(b"Hello World!\n", file_object.getvalue())


class DownloadFileToDirectoryTestCase(TestCase):
    def test_reuse_session(self) -> None:
        session = download_utils.get_session()
        with mock.patch.object(download_utils, "get_session") as session_mock:
            try:
                download_utils.download_file_to_directory(
                    download=Download(url="http://localhost", filename="dummy"),
                    directory=Path("/dummy"),
                    session=session
                )
            except Exception:
                pass
        session_mock.assert_not_called()

        with mock.patch.object(download_utils, "get_session") as session_mock:
            try:
                download_utils.download_file_to_directory(
                    download=Download(url="http://localhost", filename="dummy"),
                    directory=Path("/dummy"),
                )
            except Exception:
                pass
        session_mock.assert_called_once_with()

    def test_not_okay(self) -> None:
        session = download_utils.get_session()
        response = requests.Response()
        response.status_code = 404
        with mock.patch.object(session, "get", return_value=response):
            with self.assertRaisesRegex(
                    expected_exception=DownloadError,
                    expected_regex=r"^Download not okay\? http://localhost <Response \[404\]>$"
            ):
                download_utils.download_file_to_directory(
                    download=Download(url="http://localhost", filename="dummy"),
                    directory=Path("/dummy"),
                    session=session,
                )

    def test_hash_mismatch(self) -> None:
        session = download_utils.get_session()
        response = requests.Response()
        response.status_code = 200
        response._content = b"Hello World!\n"

        with mock.patch.object(session, "get", return_value=response):
            with self.assertRaisesRegex(
                    expected_exception=ChecksumError,
                    expected_regex=r"^Checksum mismatch: Got 03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340, expected INVALID!$"
            ):
                download_utils.download_file_to_directory(
                    download=Download(url="http://localhost", filename="dummy", sha256="INVALID"),
                    directory=Path("/dummy"),
                    session=session,
                )

    def test_valid(self) -> None:
        session = download_utils.get_session()
        response = requests.Response()
        response.status_code = 200
        response._content = b"Hello World!\n"

        with mock.patch.object(session, "get", return_value=response), TemporaryDirectory() as directory:
            download_utils.download_file_to_directory(
                download=Download(url="http://localhost", filename="test.txt", sha256="03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340"),
                directory=Path(directory),
                session=session,
            )
            self.assertEqual(
                b"Hello World!\n",
                Path(directory, "test.txt").read_bytes()
            )


class DownloadOneFilePerSecondTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.timestamps: list[datetime.datetime] = []
        self.downloads = [
            Download(url="http://localhost/file1", filename="file1.txt", sha256="03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340"),
            Download(url="http://localhost/file2", filename="file2.txt", sha256="03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340"),
            Download(url="http://localhost/file3", filename="file3.txt", sha256="03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340"),
        ]

        self.session = download_utils.get_session()
        self._session_patcher = mock.patch.object(download_utils, "get_session", return_value=self.session)
        self._session_patcher.start()
        self.addCleanup(self._session_patcher.stop)

    def test_error(self) -> None:
        # Set wrong hash.
        self.downloads[1].sha256 = "INVALID"

        def get(url: str, *args: Any, **kwargs: Any) -> requests.Response:
            response = requests.Response()
            response.status_code = 200
            response._content = b"Hello World!\n"
            return response

        with mock.patch.object(self.session, "get", side_effect=get), TemporaryDirectory() as directory:
            with self.assertRaisesRegex(
                    expected_exception=ChecksumError,
                    expected_regex=r"^Checksum mismatch: Got 03ba204e50d126e4674c005e04d82e84c21366780af1f43bd54a37816b6ab340, expected INVALID!$"
            ):
                download_utils.download_one_file_per_second(downloads=self.downloads, directory=Path(directory))
            actual = [x[1] for x in get_files_from_directory(directory)]
            self.assertEqual(["file1.txt"], actual)
            self.assertEqual(
                b"Hello World!\n",
                Path(directory, "file1.txt").read_bytes()
            )

    def test_delays(self) -> None:
        def get(url: str, *args: Any, **kwargs: Any) -> requests.Response:
            self.timestamps.append(datetime.datetime.now())
            response = requests.Response()
            response.status_code = 200
            response._content = b"Hello World!\n"
            return response

        with mock.patch.object(self.session, "get", side_effect=get), TemporaryDirectory() as directory:
            download_utils.download_one_file_per_second(downloads=self.downloads, directory=Path(directory))
            actual = [x[1] for x in get_files_from_directory(directory)]
            self.assertEqual(["file1.txt", "file2.txt", "file3.txt"], actual)
            for name in actual:
                self.assertEqual(
                    b"Hello World!\n",
                    Path(directory, name).read_bytes(),
                    name
                )

        deltas: list[datetime.timedelta] = [y - x for x, y in zip(self.timestamps[:-1], self.timestamps[1:])]
        for delta in deltas:
            self.assertGreaterEqual(delta.total_seconds(), 1)



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

"""
Tools related to pip.
"""

from __future__ import annotations

import subprocess
import sys
from dataclasses import asdict as dataclasses_asdict
from pathlib import Path

from license_tools.utils import rendering_utils

try:
    from importlib.metadata import PathDistribution
except ImportError:
    # Python < 3.8
    from importlib_metadata import PathDistribution  # type: ignore[assignment]

from piplicenses_lib import get_package_info, PackageInfo


_VERBOSE_NAMES = {
    "name": "Name",
    "version": "Version",
    "license_files": "License files",
    "author": "Author",
    "maintainer": "Maintainer",
    "license": "License",
    "license_classifiers": "License classifiers",
    "summary": "Summary",
    "homepage": "Homepage",
    "requirements": "Requirements",
}


def analyze_metadata(path: Path | str) -> PackageInfo:
    """
    Analyze the Python package metadata for the given directory.

    :param path: The directory to analyze. Should either be a `.dist-info` one or the parent.
    :return: The package metadata.
    """
    path = Path(path)
    if path.suffix not in {".dist-info", ".egg-info"}:
        try:
            path = next(path.glob("*.dist-info"))
        except StopIteration:
            path = next(path.rglob("*.egg-info"))
    distribution = PathDistribution(path)
    return get_package_info(distribution)


def check_metadata(path: Path | str) -> str:
    """
    Render the relevant details for the given package.

    :param path: The package path.
    :return: The rendered dictionary-like representation of the relevant fields.
    """
    metadata = analyze_metadata(path)
    metadata_dict = dataclasses_asdict(metadata)
    metadata_dict["license_files"] = list(metadata.license_files)
    return rendering_utils.render_dictionary(
        dictionary=metadata_dict, verbose_names_mapping=_VERBOSE_NAMES, multi_value_keys={"license_files", "license_classifiers", "requirements"}
    )


def download_package(
        package_definition: str,
        download_directory: Path | str,
        index_url: str | None = None,
        prefer_sdist: bool = False
) -> None:
    """
    Download the given package and save it to the given directory.

    :param package_definition: The Python package definition to download.
    :param download_directory: The directory to download the package to.
    :param index_url: The PyPI index URL to use. Uses the default one from the `.pypirc` file if unset.
    :param prefer_sdist: Download the source distribution instead of the wheel.
    """
    command = [
        sys.executable,
        "-m",
        "pip",
        "download",
        "--no-deps",
        package_definition,
        "--dest",
        download_directory,
    ]
    if index_url:
        command += ["--index-url", index_url]
    if prefer_sdist:
        command += ["--no-binary", ":all:"]
    try:
        subprocess.run(
            command, capture_output=True, text=True, check=True
        )
    except subprocess.CalledProcessError as exception:
        if exception.stdout:
            sys.stdout.write(exception.stdout)
        if exception.stderr:
            sys.stderr.write(exception.stderr)
        raise



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import subprocess
import sys
from contextlib import redirect_stderr, redirect_stdout
from dataclasses import asdict
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest import mock, TestCase

from license_tools.tools import pip_tools
from license_tools.utils import archive_utils
from license_tools.utils.path_utils import get_files_from_directory
from tests import get_from_url
from tests.data import JWCRYPTO__1_5_4__TAR_GZ, PYPDF__3_17_4__WHEEL


class AnalyzeMetadataTestCase(TestCase):
    def test_valid(self) -> None:
        with get_from_url(PYPDF__3_17_4__WHEEL) as path, TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            archive_utils.extract(
                archive_path=path, target_directory=directory
            )

            result1 = asdict(pip_tools.analyze_metadata(directory / "pypdf-3.17.4.dist-info"))
            result1.pop("distribution")
            result2 = asdict(pip_tools.analyze_metadata(directory))
            result2.pop("distribution")
            self.assertEqual(result1, result2)

    def test_invalid(self) -> None:
        with TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            directory.joinpath("other").mkdir()
            with self.assertRaises(StopIteration):
                pip_tools.analyze_metadata(directory)


class CheckMetadataTestCase(TestCase):
    def test_check_metadata__dist_info(self) -> None:
        with get_from_url(PYPDF__3_17_4__WHEEL) as path, TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            archive_utils.extract(
                archive_path=path, target_directory=directory
            )
            result = pip_tools.check_metadata(directory)
            self.assertEqual(f"""
               Name: pypdf
            Version: 3.17.4
      License files: {tempdir}/pypdf-3.17.4.dist-info/LICENSE
             Author: Mathieu Fenniak <biziqe@mathieu.fenniak.net>
         Maintainer: Martin Thoma <info@martin-thoma.de>
            License: UNKNOWN
License classifiers: BSD License
            Summary: A pure-python PDF library capable of splitting, merging, cropping, and transforming PDF files
           Homepage: https://github.com/py-pdf/pypdf
       Requirements:
                      * Pillow>=8.0.0 ; extra == "full"
                      * Pillow>=8.0.0 ; extra == "image"
                      * PyCryptodome ; extra == "crypto" and ( python_version == '3.6')
                      * PyCryptodome ; extra == "full" and ( python_version == '3.6')
                      * black ; extra == "dev"
                      * cryptography ; extra == "crypto" and ( python_version >= '3.7')
                      * cryptography ; extra == "full" and ( python_version >= '3.7')
                      * dataclasses; python_version < '3.7'
                      * flit ; extra == "dev"
                      * myst_parser ; extra == "docs"
                      * pip-tools ; extra == "dev"
                      * pre-commit<2.18.0 ; extra == "dev"
                      * pytest-cov ; extra == "dev"
                      * pytest-socket ; extra == "dev"
                      * pytest-timeout ; extra == "dev"
                      * pytest-xdist ; extra == "dev"
                      * sphinx ; extra == "docs"
                      * sphinx_rtd_theme ; extra == "docs"
                      * typing_extensions >= 3.7.4.3; python_version < '3.10'
                      * wheel ; extra == "dev"
"""[1:-1], result)

    def test_check_metadata__egg_info(self) -> None:
        with get_from_url(JWCRYPTO__1_5_4__TAR_GZ) as path, TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            archive_utils.extract(
                archive_path=path, target_directory=directory
            )
            result = pip_tools.check_metadata(directory)
            self.assertEqual(f"""
               Name: jwcrypto
            Version: 1.5.4
      License files: {tempdir}/jwcrypto-1.5.4/LICENSE
             Author: UNKNOWN
         Maintainer: JWCrypto Project Contributors
            License: LGPLv3+
License classifiers:
            Summary: Implementation of JOSE Web standards
           Homepage: https://github.com/latchset/jwcrypto
       Requirements:
                      * cryptography>=3.4
                      * typing_extensions>=4.5.0
"""[1:-1], result)


class DownloadPackageTestCase(TestCase):
    def test_valid_package_name(self) -> None:
        stderr = StringIO()

        with redirect_stderr(stderr), TemporaryDirectory() as directory:
            pip_tools.download_package(
                package_definition="typing_extensions==4.8.0",
                index_url="https://pypi.org/simple",
                download_directory=directory
            )
            actual = [x[1] for x in get_files_from_directory(directory)]
            self.assertEqual(
                ["typing_extensions-4.8.0-py3-none-any.whl"],
                actual
            )

        self.assertEqual("", stderr.getvalue())

    def test_prefer_source_distribution(self) -> None:
        stderr = StringIO()

        with redirect_stderr(stderr), TemporaryDirectory() as directory:
            pip_tools.download_package(
                package_definition="typing_extensions==4.8.0",
                index_url="https://pypi.org/simple",
                download_directory=directory,
                prefer_sdist=True
            )
            actual = [x[1] for x in get_files_from_directory(directory)]
            self.assertEqual(
                ["typing_extensions-4.8.0.tar.gz"],
                actual
            )

        self.assertEqual("", stderr.getvalue())

    def test_invalid_package_name(self) -> None:
        stdout, stderr = StringIO(), StringIO()

        with redirect_stdout(stdout), redirect_stderr(stderr), TemporaryDirectory() as directory:
            with self.assertRaises(subprocess.CalledProcessError):
                pip_tools.download_package(
                    package_definition="typing_extensions==1234567890",
                    index_url="https://pypi.org/simple",
                    download_directory=directory
                )
            actual = [x[1] for x in get_files_from_directory(directory)]
            self.assertEqual(
                [],
                actual
            )

        stderr_string = stderr.getvalue()
        self.assertEqual("", stdout.getvalue())
        self.assertIn(
            "ERROR: Could not find a version that satisfies the requirement typing_extensions==1234567890 (from versions: ",
            stderr_string,
        )
        self.assertIn(
            "\nERROR: No matching distribution found for typing_extensions==1234567890\n",
            stderr_string,
        )

    def test_index_url_handling(self) -> None:
        directories = []

        def run(
            command: list[str | Path], *args: Any, **kwargs: Any
        ) -> "subprocess.CompletedProcess[bytes]":
            _directory = command[command.index("--dest") + 1]
            directories.append(_directory)
            Path(_directory).joinpath("dummy.py").touch()
            return subprocess.CompletedProcess(args=args, returncode=0, stdout=b"")

        with mock.patch("subprocess.run", side_effect=run) as subprocess_mock:
            with TemporaryDirectory() as directory:
                pip_tools.download_package(
                    package_definition="testing", download_directory=directory
                )
            with TemporaryDirectory() as directory:
                pip_tools.download_package(
                    package_definition="testing",
                    index_url="DUMMY",
                    download_directory=directory,
                )

            subprocess_mock.assert_has_calls(
                [
                    mock.call(
                        [
                            sys.executable,
                            "-m",
                            "pip",
                            "download",
                            "--no-deps",
                            "testing",
                            "--dest",
                            directories[0],
                        ],
                        capture_output=True,
                        text=True,
                        check=True,
                    ),
                    mock.call(
                        [
                            sys.executable,
                            "-m",
                            "pip",
                            "download",
                            "--no-deps",
                            "testing",
                            "--dest",
                            directories[1],
                            "--index-url",
                            "DUMMY",
                        ],
                        capture_output=True,
                        text=True,
                        check=True,
                    ),
                ],
                any_order=False,
            )
            self.assertEqual(
                2, subprocess_mock.call_count, subprocess_mock.call_args_list
            )



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

import copy
import os
import re
import tarfile
from contextlib import contextmanager, redirect_stdout
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Any, cast, Dict, Generator, List, Set, Tuple  # TODO: Remove `Set` and `Tuple` after dropping Python 3.8.
from unittest import mock, TestCase

from license_tools import retrieval
from license_tools.retrieval import RetrievalFlags
from license_tools.tools.scancode_tools import FileResults, LicenseDetection, LicenseMatch, Licenses
from license_tools.utils.path_utils import get_files_from_directory
from tests import Download, get_from_url
from tests.data import (
    BASE64__0_22_0__CARGO_TOML,
    LIBAIO1__0_3_109_1_25__RPM,
    SETUP_PATH,
    SETUP_PY_LICENSES,
    TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT,
    TYPING_EXTENSION_4_8_0__LICENSES,
    TYPING_EXTENSION_4_8_0__SOURCE_FILES,
    TYPING_EXTENSION_4_8_0__WHEEL_FILES,
    TYPING_EXTENSIONS__4_8_0__SDIST,
    TYPING_EXTENSIONS__4_8_0__WHEEL,
)


class RetrievalFlagsTestCase(TestCase):
    def test_to_int(self) -> None:
        self.assertEqual(0, RetrievalFlags.to_int())
        self.assertEqual(
            21, RetrievalFlags.to_int(True, False, True, False, True, False)
        )

    def test_all(self) -> None:
        self.assertEqual(255, RetrievalFlags.all())
        self.assertDictEqual(
            dict(
                retrieve_copyrights=True,
                retrieve_emails=True,
                retrieve_file_info=True,
                retrieve_urls=True,
                retrieve_ldd_data=True,
                retrieve_font_data=True,
                retrieve_python_metadata=True,
                retrieve_cargo_metadata=True,
            ),
            cast(Dict[str, bool], RetrievalFlags.all(as_kwargs=True)),
        )

    def test_is_set(self) -> None:
        self.assertFalse(RetrievalFlags.is_set(flags=0, flag=RetrievalFlags.EMAILS))
        self.assertTrue(RetrievalFlags.is_set(flags=2, flag=RetrievalFlags.EMAILS))
        self.assertTrue(RetrievalFlags.is_set(flags=31, flag=RetrievalFlags.EMAILS))
        self.assertFalse(RetrievalFlags.is_set(flags=9, flag=RetrievalFlags.EMAILS))

    def test_to_kwargs(self) -> None:
        self.assertDictEqual(
            dict(
                retrieve_copyrights=False,
                retrieve_emails=False,
                retrieve_file_info=False,
                retrieve_urls=False,
                retrieve_ldd_data=False,
                retrieve_font_data=False,
                retrieve_python_metadata=False,
                retrieve_cargo_metadata=False,
            ),
            RetrievalFlags.to_kwargs(0),
        )
        self.assertDictEqual(
            dict(
                retrieve_copyrights=True,
                retrieve_emails=False,
                retrieve_file_info=True,
                retrieve_urls=False,
                retrieve_ldd_data=True,
                retrieve_font_data=False,
                retrieve_python_metadata=False,
                retrieve_cargo_metadata=False,
            ),
            RetrievalFlags.to_kwargs(21),
        )


class RunOnFileTestCase(TestCase):
    def _run_mocked(
        self,
        flags: int,
        mock_target: str,
        return_value: str | None = "",
    ) -> tuple[mock.Mock, mock.Mock, str]:
        stdout = StringIO()

        class DummyFileResult:
            licenses = None

            def __init__(self, *args: Any, **kwargs: Any) -> None:
                pass

        file_result = DummyFileResult()
        with mock.patch.object(
            retrieval, "FileResults", return_value=file_result
        ) as results_mock, redirect_stdout(stdout), mock.patch(
            mock_target, return_value=return_value
        ) as check_mock:
            result = retrieval.run_on_file(
                path=SETUP_PATH, short_path="setup.py", retrieval_flags=flags
            )
            self.assertEqual(file_result, result)
        return results_mock, check_mock, stdout.getvalue()

    def test_run_on_file__ldd_handling(self) -> None:
        # 1) LDD handling is inactive.
        results_mock, check_mock, stdout = self._run_mocked(
            flags=15, mock_target="license_tools.tools.linking_tools.check_shared_objects"
        )
        check_mock.assert_not_called()
        results_mock.assert_called_once_with(
            path=SETUP_PATH,
            short_path="setup.py",
            retrieve_licenses=True,
            retrieve_copyrights=True,
            retrieve_emails=True,
            retrieve_file_info=True,
            retrieve_urls=True,
        )
        self.assertEqual("", stdout)

        # 2) LDD handling is active, but has no results.
        for result in ["", None]:
            with self.subTest(result=result):
                results_mock, check_mock, stdout = self._run_mocked(
                    flags=31,
                    return_value=result,
                    mock_target="license_tools.tools.linking_tools.check_shared_objects",
                )
                check_mock.assert_called_once_with(path=SETUP_PATH)
                results_mock.assert_called_once_with(
                    path=SETUP_PATH,
                    short_path="setup.py",
                    retrieve_licenses=True,
                    retrieve_copyrights=True,
                    retrieve_emails=True,
                    retrieve_file_info=True,
                    retrieve_urls=True,
                )
                self.assertEqual("", stdout)

        # 3) LDD handling is active and has results.
        ldd_usr_bin_bc = """    linux-vdso.so.1 (0x00007fff30abf000)
    libreadline.so.7 => /lib64/libreadline.so.7 (0x00007fbe48c00000)
    libc.so.6 => /lib64/libc.so.6 (0x00007fbe48a09000)
    libtinfo.so.6 => /lib64/libtinfo.so.6 (0x00007fbe48600000)
    /lib64/ld-linux-x86-64.so.2 (0x00007fbe492b8000)
"""
        results_mock, check_mock, stdout = self._run_mocked(
            flags=31,
            return_value=ldd_usr_bin_bc,
            mock_target="license_tools.tools.linking_tools.check_shared_objects",
        )
        check_mock.assert_called_once_with(path=SETUP_PATH)
        results_mock.assert_called_once_with(
            path=SETUP_PATH,
            short_path="setup.py",
        )
        self.assertEqual("setup.py\n" + ldd_usr_bin_bc + "\n", stdout)

    def test_run_on_file__font_handling(self) -> None:
        # 1) Font handling is inactive.
        results_mock, check_mock, stdout = self._run_mocked(
            flags=15, mock_target="license_tools.tools.font_tools.check_font"
        )
        check_mock.assert_not_called()
        results_mock.assert_called_once_with(
            path=SETUP_PATH,
            short_path="setup.py",
            retrieve_licenses=True,
            retrieve_copyrights=True,
            retrieve_emails=True,
            retrieve_file_info=True,
            retrieve_urls=True,
        )
        self.assertEqual("", stdout)

        # 2) Font handling is active, but has no results.
        for result in ["", None]:
            with self.subTest(result=result):
                results_mock, check_mock, stdout = self._run_mocked(
                    flags=63,
                    return_value=result,
                    mock_target="license_tools.tools.font_tools.check_font",
                )
                check_mock.assert_called_once_with(path=SETUP_PATH)
                results_mock.assert_called_once_with(
                    path=SETUP_PATH,
                    short_path="setup.py",
                    retrieve_licenses=True,
                    retrieve_copyrights=True,
                    retrieve_emails=True,
                    retrieve_file_info=True,
                    retrieve_urls=True,
                )
                self.assertEqual("", stdout)

        # 3) Font handling is active and has results.
        font_awesome = """             Copyright notice: Copyright (c) Font Awesome
          Font family name: Font Awesome 6 Free Solid
       Font subfamily name: Solid
    Unique font identifier: Font Awesome 6 Free Solid-6.5.1
            Full font name: Font Awesome 6 Free Solid
            Version string: Version 773.01171875 (Font Awesome version: 6.5.1)
           PostScript name: FontAwesome6Free-Solid
               Description: The web's most popular icon set and toolkit.
                URL Vendor: https://fontawesome.com
   Typographic Family name: Font Awesome 6 Free
Typographic Subfamily name: Solid
"""
        results_mock, check_mock, stdout = self._run_mocked(
            flags=63,
            return_value=font_awesome,
            mock_target="license_tools.tools.font_tools.check_font",
        )
        check_mock.assert_called_once_with(path=SETUP_PATH)
        results_mock.assert_called_once_with(
            path=SETUP_PATH,
            short_path="setup.py",
        )
        self.assertEqual("setup.py\n" + font_awesome + "\n\n", stdout)

    def test_cargo_toml(self) -> None:
        with get_from_url(BASE64__0_22_0__CARGO_TOML) as source_path, TemporaryDirectory() as directory:
            cargo_toml_path = Path(directory) / "Cargo.toml"
            cargo_toml_path.write_bytes(source_path.read_bytes())

            stdout = StringIO()
            with redirect_stdout(stdout):
                retrieval.run_on_file(path=cargo_toml_path, short_path="/path/to/Cargo.toml")
            self.assertEqual("", stdout.getvalue())

            stdout = StringIO()
            with redirect_stdout(stdout):
                result = retrieval.run_on_file(path=cargo_toml_path, short_path="/path/to/Cargo.toml", retrieval_flags=RetrievalFlags.CARGO_METADATA)

            self.assertEqual(
                """
/path/to/Cargo.toml
        Name: base64
     Version: 0.22.0
     Authors:
               * Alice Maz <alice@alicemaz.com>
               * Marshall Pierce <marshall@mpierce.org>
 Description: encodes and decodes base64 as bytes or utf8
      README: README.md
  Repository: https://github.com/marshallpierce/rust-base64
     License: MIT OR Apache-2.0
    Keywords:
               * base64
               * decode
               * encode
               * no_std
               * utf8
  Categories: encoding

"""[1:],
                stdout.getvalue()
            )
            self.assertEqual(
                FileResults(
                    path=cargo_toml_path, short_path='/path/to/Cargo.toml', retrieve_copyrights=False, retrieve_emails=False, retrieve_urls=False,
                    retrieve_licenses=True, retrieve_file_info=False, copyrights=None, emails=None, urls=None,
                    licenses=Licenses(
                        detected_license_expression='mit OR apache-2.0', detected_license_expression_spdx='MIT OR Apache-2.0',
                        percentage_of_license_text=2.55,
                        license_detections=[
                            LicenseDetection(
                                license_expression='mit OR apache-2.0', license_expression_spdx='MIT OR Apache-2.0',
                                identifier='mit_or_apache_2_0-719f8427-422e-8023-c20e-9f8dd0af13b9',
                                matches=[
                                    LicenseMatch(
                                        score=100.0, start_line=11, end_line=11, matched_length=6, match_coverage=100.0, matcher='2-aho',
                                        license_expression='mit OR apache-2.0', spdx_license_expression='MIT OR Apache-2.0',
                                        rule_identifier='mit_or_apache-2.0_14.RULE', rule_relevance=100,
                                        rule_url='https://github.com/nexB/scancode-toolkit/tree/develop/src/licensedcode/data/rules/mit_or_apache-2.0_14.RULE',
                                        from_file=None, matched_text=None
                                    )
                                ]
                            )
                        ],
                        license_clues=[]
                    ),
                    file_info=None
                ),
                result
            )


class RunOnDirectoryTestCase(TestCase):
    def test_run_on_directory(self) -> None:
        file_results = [object()] * 5
        file_results_iterable = iter(file_results)
        paths = [(Path(f"/tmp/file{i}.py"), f"file{i}.py") for i in range(1, 6)]

        def run_on_file(path: Path, short_path: str, retrieval_flags: int = 0) -> Any:
            return next(file_results_iterable)

        with mock.patch.object(
            retrieval, "run_on_file", side_effect=run_on_file
        ) as run_mock, mock.patch.object(
            retrieval, "get_files_from_directory", return_value=paths
        ) as get_mock:
            results = list(
                retrieval.run_on_directory(
                    "/tmp/dummy/directory", job_count=1, retrieval_flags=42
                )
            )
        self.assertListEqual(file_results, results)
        run_mock.assert_has_calls(
            [
                mock.call(
                    path=current_path, short_path=current_short_path, retrieval_flags=42
                )
                for current_path, current_short_path in paths
            ],
            any_order=False,
        )
        self.assertEqual(len(paths), run_mock.call_count, run_mock.call_args_list)
        get_mock.assert_called_once_with("/tmp/dummy/directory", None)

    @classmethod
    def _generate_tar_bz2_archive(cls, directory: Path) -> Path:
        with TemporaryDirectory() as nested_directory:
            nested_path = Path(nested_directory)
            nested_path.joinpath("LICENSE").write_text("This is my license.")
            with tarfile.open(directory / "nested.tar.bz2", "w:bz2") as tar:
                tar.add(nested_path, arcname=nested_path.name)
        return nested_path

    def _test_nested(
            self, result_set: Set[Path], directory: Path, nested_path: Path, expected: List[Tuple[Path, str]], run_mock: mock.Mock
    ) -> None:
        self.assertEqual(1, len(result_set), result_set)
        remaining = result_set.pop()
        self.assertEqual(directory, remaining.parent.parent.parent)
        self.assertEqual(nested_path.name, remaining.parent.name)
        self.assertEqual("LICENSE", remaining.name)
        expected.append((remaining, "/".join(str(remaining).rsplit("/", maxsplit=3)[1:])))

        run_mock.assert_has_calls(
            [
                mock.call(
                    path=current_path, short_path=current_short_path, retrieval_flags=42
                )
                for current_path, current_short_path in expected
            ],
            any_order=False,
        )
        self.assertEqual(len(expected), run_mock.call_count, run_mock.call_args_list)

    def test_nested_with_existing_directory(self) -> None:
        with TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            directory.joinpath("directory").mkdir()
            directory.joinpath("directory", "file.txt").write_text("MIT-0")
            directory.joinpath("nested_tar_bz2").write_text("Dummy")
            nested_path = self._generate_tar_bz2_archive(directory)

            def run_on_file(path: Path, short_path: str, retrieval_flags: int = 0) -> Any:
                return path

            with mock.patch.object(
                retrieval, "run_on_file", side_effect=run_on_file
            ) as run_mock:
                results = list(
                    retrieval.run_on_directory(
                        tempdir, job_count=1, retrieval_flags=42
                    )
                )

            self.assertSetEqual(
                {"directory", "nested_tar_bz2", "nested.tar.bz2"},
                {path.name for path in directory.glob("*")}
            )

        result_set: Set[Path] = cast(Set[Path], set(results))
        expected: List[Tuple[Path, str]] = []
        self.assertEqual(4, len(results), results)
        for name in ["directory/file.txt", "nested.tar.bz2", "nested_tar_bz2"]:
            result_set.remove(directory / name)
            expected.append((directory / name, name))
        self._test_nested(
            result_set=result_set, directory=directory, nested_path=nested_path, expected=expected, run_mock=run_mock
        )

    def test_nested_without_existing_directory(self) -> None:
        with TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            directory.joinpath("directory").mkdir()
            directory.joinpath("directory", "file.txt").write_text("MIT-0")
            nested_path = self._generate_tar_bz2_archive(directory)

            def run_on_file(path: Path, short_path: str, retrieval_flags: int = 0) -> Any:
                return path

            with mock.patch.object(
                retrieval, "run_on_file", side_effect=run_on_file
            ) as run_mock:
                results = list(
                    retrieval.run_on_directory(
                        tempdir, job_count=1, retrieval_flags=42
                    )
                )

            self.assertSetEqual(
                {"directory", "nested.tar.bz2"},
                {path.name for path in directory.glob("*")}
            )

        result_set: Set[Path] = cast(Set[Path], set(results))
        expected: List[Tuple[Path, str]] = []
        self.assertEqual(3, len(results), results)
        for name in ["directory/file.txt", "nested.tar.bz2"]:
            result_set.remove(directory / name)
            expected.append((directory / name, name))
        self._test_nested(
            result_set=result_set, directory=directory, nested_path=nested_path, expected=expected, run_mock=run_mock
        )

    def test_nested_storage_variant(self) -> None:
        with TemporaryDirectory() as tempdir:
            directory = Path(tempdir)
            directory.joinpath("directory").mkdir()
            directory.joinpath("directory", "file.txt").write_text("MIT-0")
            _ = self._generate_tar_bz2_archive(directory)

            def run_on_file(path: Path, short_path: str, retrieval_flags: int = 0) -> Any:
                return path

            with mock.patch.object(retrieval, "run_on_file", side_effect=run_on_file):
                list(
                    retrieval.run_on_directory(
                        tempdir, job_count=1, retrieval_flags=42,
                        allow_random_directory_for_archive=False,
                        delete_unpacked_archive_directories=False,
                    )
                )

            self.assertSetEqual(
                {"directory", "nested.tar.bz2", "nested_tar_bz2"},
                {path.name for path in directory.glob("*")}
            )

            with mock.patch.object(retrieval, "run_on_file", side_effect=run_on_file):
                with self.assertRaisesRegex(
                        expected_exception=FileExistsError,
                        expected_regex=fr"^\[Errno 17\] File exists: '{re.escape(str(directory / 'nested_tar_bz2'))}'$"
                ):
                    list(
                        retrieval.run_on_directory(
                            tempdir, job_count=1, retrieval_flags=42,
                            allow_random_directory_for_archive=False,
                            delete_unpacked_archive_directories=False,
                        )
                    )


class RunOnPackageArchiveFileTestCase(TestCase):
    def _check_call(self, download: Download, expected_files: List[str], expected_license: str | None = None) -> None:
        with get_from_url(download) as archive_path:
            directory_result = [object(), object(), object()]

            def run_on_directory(
                directory: Path, job_count: int, retrieval_flags: int
            ) -> Generator[Any, None, None]:
                self.assertEqual(2, job_count)
                self.assertEqual(42, retrieval_flags)
                actual = [x[1] for x in get_files_from_directory(directory)]
                self.assertListEqual(expected_files, actual)
                yield from directory_result

            with mock.patch.object(
                retrieval, "run_on_directory", side_effect=run_on_directory
            ):
                result = list(
                    retrieval.run_on_package_archive_file(
                        archive_path=archive_path, job_count=2, retrieval_flags=42
                    )
                )

            expected = directory_result
            if expected_license:
                dummy_entry = retrieval._get_dummy_file_results(path=archive_path, short_path=archive_path.name)
                dummy_entry.licenses = Licenses(
                    detected_license_expression=expected_license,
                    detected_license_expression_spdx=expected_license,
                )
                expected.insert(0, dummy_entry)
            self.assertEqual(expected, result)

    def test_wheel_file(self) -> None:
        self._check_call(
            download=TYPING_EXTENSIONS__4_8_0__WHEEL, expected_files=TYPING_EXTENSION_4_8_0__WHEEL_FILES,
        )

    def test_non_wheel_file(self) -> None:
        self._check_call(
            download=TYPING_EXTENSIONS__4_8_0__SDIST, expected_files=TYPING_EXTENSION_4_8_0__SOURCE_FILES,
        )

    def test_rpm(self) -> None:
        self._check_call(
            download=LIBAIO1__0_3_109_1_25__RPM,
            expected_files=[
                "lib64/libaio.so.1.0.1",
                "usr/share/doc/packages/libaio1/COPYING",
                "usr/share/doc/packages/libaio1/TODO",
            ],
            expected_license="LGPL-2.1-or-later",
        )


class RunOnDownloadedArchiveFileTestCase(TestCase):
    def _check_call(self, download: Download) -> None:
        directory_result = [object(), object(), object()]

        def run_on_package_archive_file(
            archive_path: Path, job_count: int, retrieval_flags: int
        ) -> Generator[Any, None, None]:
            self.assertEqual(2, job_count)
            self.assertEqual(42, retrieval_flags)
            self.assertEqual(download.suffix, archive_path.name[-len(download.suffix):])
            yield from directory_result

        with mock.patch.object(
            retrieval,
            "run_on_package_archive_file",
            side_effect=run_on_package_archive_file,
        ):
            result = list(
                retrieval.run_on_downloaded_archive_file(
                    download_url=download.url, job_count=2, retrieval_flags=42
                )
            )
        self.assertEqual(directory_result, result)

    def test_wheel_file(self) -> None:
        self._check_call(download=TYPING_EXTENSIONS__4_8_0__WHEEL)

    def test_non_wheel_file(self) -> None:
        self._check_call(download=TYPING_EXTENSIONS__4_8_0__SDIST)


class RunOnDownloadedPackageFileTestCase(TestCase):
    def test_valid_package_name(self) -> None:
        archive_result = [object(), object(), object()]

        def run_on_package_archive_file(
            archive_path: Path, job_count: int, retrieval_flags: int, retrieve_python_metadata: bool = False
        ) -> Generator[Any, None, None]:
            self.assertEqual("typing_extensions-4.8.0-py3-none-any.whl", archive_path.name)
            self.assertEqual(3, job_count)
            self.assertEqual(42, retrieval_flags)
            self.assertEqual(31584, len(archive_path.read_bytes()))
            yield from archive_result

        with mock.patch.object(
            retrieval,
            "run_on_package_archive_file",
            side_effect=run_on_package_archive_file,
        ):
            result = list(
                retrieval.run_on_downloaded_package_file(
                    package_definition="typing_extensions==4.8.0",
                    index_url="https://pypi.org/simple",
                    job_count=3,
                    retrieval_flags=42,
                )
            )
            self.assertEqual(archive_result, result)

    def test_prefer_source_distribution(self) -> None:
        archive_result = [object(), object(), object()]

        def run_on_package_archive_file(
            archive_path: Path, job_count: int, retrieval_flags: int, retrieve_python_metadata: bool = False
        ) -> Generator[Any, None, None]:
            self.assertEqual("typing_extensions-4.8.0.tar.gz", archive_path.name)
            self.assertEqual(3, job_count)
            self.assertEqual(42, retrieval_flags)
            self.assertEqual(71456, len(archive_path.read_bytes()))
            yield from archive_result

        with mock.patch.object(
            retrieval,
            "run_on_package_archive_file",
            side_effect=run_on_package_archive_file,
        ):
            result = list(
                retrieval.run_on_downloaded_package_file(
                    package_definition="typing_extensions==4.8.0",
                    index_url="https://pypi.org/simple",
                    job_count=3,
                    retrieval_flags=42,
                    prefer_sdist=True,
                )
            )
            self.assertEqual(archive_result, result)


class CheckThatExactlyOneValueIsSetTestCase(TestCase):
    def test_check_that_exactly_one_value_is_set(self) -> None:
        self.assertIs(
            True,
            retrieval._check_that_exactly_one_value_is_set([None, None, "test", None]),
        )
        self.assertIs(False, retrieval._check_that_exactly_one_value_is_set([]))
        self.assertIs(
            False, retrieval._check_that_exactly_one_value_is_set([None, None])
        )


class Stdout:
    def __init__(self) -> None:
        self.stdout = StringIO()

    def __str__(self) -> str:
        return self.stdout.getvalue()


class RunTestCase(TestCase):
    @contextmanager
    def record_stdout(self) -> Generator[Stdout, None, None]:
        result = Stdout()
        with mock.patch(
            "shutil.get_terminal_size", return_value=os.terminal_size((100, 20))
        ), redirect_stdout(result.stdout):
            yield result

    def test_package_definition(self) -> None:
        with self.record_stdout() as stdout:
            with mock.patch.object(
                retrieval,
                "run_on_downloaded_package_file",
                return_value=iter(TYPING_EXTENSION_4_8_0__LICENSES),
            ) as run_mock:
                result = retrieval.run(
                    package_definition="typing_extensions==4.8.0",
                    index_url="https://example.org/simple",
                )
        run_mock.assert_called_once_with(
            package_definition="typing_extensions==4.8.0",
            index_url="https://example.org/simple",
            retrieval_flags=0,
            job_count=4,
            prefer_sdist=False,
        )
        self.assertEqual(TYPING_EXTENSION_4_8_0__LICENSES, result)
        self.assertEqual(TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT, str(stdout))

    def test_package_definition__with_metadata(self) -> None:
        self.maxDiff = None
        with self.record_stdout() as stdout:
            result = retrieval.run(
                package_definition="typing_extensions==4.8.0",
                index_url="https://pypi.org/simple",
                retrieve_python_metadata=True,
            )

        expected_result = copy.deepcopy(TYPING_EXTENSION_4_8_0__LICENSES)
        for entry in expected_result:
            entry.path = Path("dummy")
            entry.retrieve_licenses = True
        for entry in result:
            entry.path = Path("dummy")
        self.assertEqual(expected_result, result)

        expected_output = """               Name: typing_extensions
            Version: 4.8.0
      License files: /tmp/dummy/typing_extensions-4.8.0.dist-info/LICENSE
             Author: "Guido van Rossum, Jukka Lehtosalo, Łukasz Langa, Michael Lee" <levkivskyi@gmail.com>
         Maintainer: UNKNOWN
            License: UNKNOWN
License classifiers: Python Software Foundation License
            Summary: Backported and Experimental Type Hints for Python 3.8+
           Homepage: https://github.com/python/typing_extensions
       Requirements:

""" + TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT
        self.assertEqual(expected_output, re.sub(pattern=r"/tmp/tmp[^/]+", repl="/tmp/dummy", string=str(stdout)))

    def test_directory(self) -> None:
        with TemporaryDirectory() as directory:
            path = Path(directory)
            with self.record_stdout() as stdout:
                with mock.patch.object(
                    retrieval,
                    "run_on_directory",
                    return_value=iter(TYPING_EXTENSION_4_8_0__LICENSES),
                ) as run_mock:
                    result = retrieval.run(directory=path, retrieve_ldd_data=True)
            run_mock.assert_called_once_with(
                directory=directory, retrieval_flags=16, job_count=4
            )
            self.assertEqual(TYPING_EXTENSION_4_8_0__LICENSES, result)
            self.assertEqual(TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT, str(stdout))

    def test_archive_path(self) -> None:
        with self.record_stdout() as stdout:
            with mock.patch.object(
                retrieval,
                "run_on_package_archive_file",
                return_value=iter(TYPING_EXTENSION_4_8_0__LICENSES),
            ) as run_mock:
                result = retrieval.run(
                    archive_path=Path("/tmp/dummy/typing_extensions-4.8.0.tar.gz"),
                    retrieve_copyrights=True,
                    job_count=1,
                )
        run_mock.assert_called_once_with(
            archive_path=Path("/tmp/dummy/typing_extensions-4.8.0.tar.gz"),
            retrieval_flags=1,
            job_count=1,
        )
        self.assertEqual(TYPING_EXTENSION_4_8_0__LICENSES, result)
        self.assertEqual(TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT, str(stdout))

    def test_download_url(self) -> None:
        with self.record_stdout() as stdout:
            with mock.patch.object(
                retrieval,
                "run_on_downloaded_archive_file",
                return_value=iter(TYPING_EXTENSION_4_8_0__LICENSES),
            ) as run_mock:
                result = retrieval.run(
                    download_url="https://example.org/archive.tar.gz",
                    retrieve_copyrights=True,
                    job_count=1,
                )
        run_mock.assert_called_once_with(
            download_url="https://example.org/archive.tar.gz",
            retrieval_flags=1,
            job_count=1,
        )
        self.assertEqual(TYPING_EXTENSION_4_8_0__LICENSES, result)
        self.assertEqual(TYPING_EXTENSION_4_8_0__EXPECTED_OUTPUT, str(stdout))

    def test_file_path(self) -> None:
        with self.record_stdout() as stdout:
            result = retrieval.run(file_path=SETUP_PATH, job_count=1)
        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result), result)
        first_result = result[0]
        self.assertIsInstance(first_result, FileResults)
        self.assertEqual(SETUP_PY_LICENSES, first_result.licenses)
        self.assertEqual(
            f"""{SETUP_PATH} Apache-2.0 AND (LicenseRef-scancode-unknown-license-reference AND Apache-2.0) 

====================================================================================================

Apache-2.0 AND (LicenseRef-scancode-unknown-license-reference AND Apache-2.0)  1
""",  # noqa: W291
            str(stdout),
        )

    def test_nested_archive(self) -> None:
        # url = "https://download.opensuse.org/source/distribution/leap/15.6/repo/oss/src/libaio-0.3.109-1.25.src.rpm"  # Takes too long.
        with TemporaryDirectory() as archive_directory, NamedTemporaryFile(suffix=".tar.gz") as archive_file:
            archive_directory_path = Path(archive_directory)
            archive_directory_path.joinpath("directory1", "directory2").mkdir(parents=True)
            archive_directory_path.joinpath("directory1", "directory2", "file.txt").write_text("MIT-0")
            archive_directory_path.joinpath("directory1", "directory2", "file2.txt").write_text("Apache-2.0")
            with TemporaryDirectory() as nested_directory:
                nested_path = Path(nested_directory)
                nested_path.joinpath("subdirectory").mkdir()
                nested_path.joinpath("subdirectory", "README").write_text("CC-BY-2.0")
                nested_path.joinpath("LICENSE").write_text("This is my license.")
                with tarfile.open(archive_directory_path / "nested.tar.bz2", "w:bz2") as tar:
                    tar.add(nested_path, arcname=nested_path.name)

            archive_path = Path(archive_file.name)
            with tarfile.open(archive_path, "w:gz") as tar:
                tar.add(archive_directory_path, arcname=archive_directory_path.name)

            with self.record_stdout() as stdout:
                result = retrieval.run(archive_path=archive_path, job_count=1)

        self.assertIsInstance(result, list)
        self.assertEqual(5, len(result), result)
        self.assertEqual(
            f"""                {archive_directory_path.name}/directory1/directory2/file.txt                                                                        
               {archive_directory_path.name}/directory1/directory2/file2.txt                                                             Apache-2.0 [100.0]
                                {archive_directory_path.name}/nested.tar.bz2                                                                        
            {archive_directory_path.name}/nested_tar_bz2/{nested_path.name}/LICENSE                                                                        
{archive_directory_path.name}/nested_tar_bz2/{nested_path.name}/subdirectory/README                                                              CC-BY-2.0 [50.0]

====================================================================================================

                                                            Apache-2.0  1
                                                             CC-BY-2.0  1
                                                                  None  3
""",  # noqa: E501, W291
            str(stdout),
        )



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from __future__ import annotations

from unittest import TestCase

from license_tools import constants


class ConstantsTestCase(TestCase):
    def test_version(self) -> None:
        self.assertIsNotNone(constants.VERSION)



# Copyright (c) stefan6419846. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.

from pathlib import Path

import setuptools


ROOT_DIRECTORY = Path(__file__).parent.resolve()


setuptools.setup(
    name="license_tools",
    description="Collection of tools for working with Open Source licenses",
    version="0.9.0",
    license="Apache-2.0",
    long_description=Path(ROOT_DIRECTORY / "README.md").read_text(encoding="UTF-8"),
    long_description_content_type="text/markdown",
    author="stefan6419846",
    url="https://github.com/stefan6419846/license_tools",
    packages=setuptools.find_packages(include=["license_tools", "license_tools.*"]),
    include_package_data=True,
    python_requires=">=3.8, <4",
    install_requires=[
        "scancode-toolkit>=32.1.0",
        "typecode-libmagic",
        "joblib",
        "fontTools[woff]",
        "pip-licenses-lib",
        "requests",
        "tomli",
    ],
    extras_require={
        "dev": [
            "codespell",
            "coverage",
            "flake8",
            "flake8-bugbear",
            "pep8-naming",
            "mypy",
            "types-requests",
            "importlib_resources; python_version<'3.9'",
        ],
        "extended_rpm": [
            "rpmfile",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    keywords=["open source", "license", "package", "dependency", "licensing"],
)
